;******************************************************************************
;*                                                                            *
;*                            License and T&C                                 *
;*                                                                            *
;*        GR8BIT storage subsystem microcontroller microcode version 1.0      *
;*             This software is copyleft 2011 by Eugeny Brychkov              *
;*                                                                            *
;* You can use and modify its source and executable, given two conditions are *
;* met: (a) you keep this Copyleft statement in the source code, and you do   *
;* not distribute software and resulting executable on the commercial basis   *
;* and on form the purposes, not related to GR8BIT platform, without written  *
;* permission of Eugeny Brychkov (eugeny_brychkov@hotmail.com)                *
;*                                                                            *
;******************************************************************************

; version 1.0 (Nov 28, 2011)	- initial release
; version 1.0a (Jan 26, 2012)	- Storage subsystem ROM was released, this code was intact
; version 1.0b (Mar 14, 2012)	- added support for 720K media, fixed some bugs regarding
;								  handling of i/o errors and disk removals, improved
;								  disk change handling. No retry algorithm is implemented
;								  so far. If r/w operation fails and both speeds are tried,
;								  execution is terminated
; version 2.0 (Apr 02, 2012)	- initial version for PIC18F46K22, almost entirely discarded
;								  handshaking alorithm, added VERSION command
; version 2.0a (May 10, 2012)	- complete remake of all the I/O code; FDC now uses DMA;
;								  implemented HDD routines; FDD retry algorithm now is in
;								  MC; developed MAKEDPB and other routines; DSKIO now has
;								  23-bit access

;///////////////////////////////////////////////////////////////////////////////
;
;							TECHNICAL INFORMATION
; GR8BIT storage subsystem based on PIC18F46K22 does not require additional
; wait states at the 3.579545MHz CPU speed
;
; The principle of operation of the storage subsystem
; There're two ports in GR8BIT, &hA4 (port base+0) and &hA5 (port base+1), which
; control storage subsystem microcontroller. Port base+0 is used for command
; initiation and abortion (in some cases it is not possible to abort command in
; execution). Port base+1 is used for light-weight handshaking and data transfers.
;
; Microcontroller is running at 64MHz system clock (62.5ns instruction cycle)
;
;///////////////////////////////////////////////////////////////////////////////

; GAP lengths can be found in SMC37C65 data sheet
; Info on DSPEC values can be found here http://wiki.osdev.org/Floppy_Disk_Controller, also http://www.isdaman.com/alsos/hardware/fdc/floppy.htm
; and also http://www.mathematik.uni-kl.de/~hannes/en/floppy.diff

RETRIES		equ		0x04

; 1.44MB diskette / speed configuration
GPL1_144	equ		0x1B		; gap length when reading
GPL2_144	equ		0x54		; gap length when formatting
DSPEC1_144	equ		0xAF		; A=step rate 6ms, F=head unload delay (240ms)
DSPEC2_144	equ		0x11		; 8=16ms head load time, 1=no DMA
DSPD_144	equ		0x0			; speed 500kB/s
SpT_144		equ		0x12		; sectors per track (18)

; 720K diskette / speed configuration
GPL1_720	equ	0x2A
GPL2_720	equ	0x50
DSPEC1_720	equ	0xDF		; D=step rate 6ms, F=head unload time (480ms)
DSPEC2_720	equ	0x11		; 8=32ms head load time (very safe), 1=no DMA
DSPD_720	equ	0x02		; speed 250kB/s (CR1=1, CR0=0, DRV=x)
SpT_720		equ	0x09		; sectors per track (9)

	list		p=18f46k22			; list directive to define processor
	#include	<P18F46K22.inc>		; processor specific variable definitions`

; Microcontroller configuration
	CONFIG	FOSC=INTIO67		; internal oscillator, RA6 and RA7 are generic I/O pins
	CONFIG	PLLCFG=ON			; turn PLL (4*) on
	CONFIG	FCMEN=OFF			; fail-safe clock monitor off
	CONFIG	IESO=OFF			; internal/external oscillator switching is off
	CONFIG	PWRTEN=OFF			; power up timer off (using internal oscillator only)
	CONFIG	BOREN=OFF			; brown-out reset disabled completely
	CONFIG	WDTEN=OFF			; watchdog timer is controlled by SWDTEN bit
	CONFIG	PBADEN=OFF			; port B bits 4:0 are digital I/O
	CONFIG	MCLRE=INTMCLR		; MCRLE function is off (pin is used as digital input)
	CONFIG	LVP=OFF
	CONFIG	STVREN=OFF

; Command buffer memory organization
;	output: ACCESS bank, sequence of the bytes (or words) to send to controller (offset 0x00-0x0F)
;	intput: ACCESS bank, sequence of the bytes (or words) read from the controller (offset 0x10-0x1F)
CMDWLOC		equ	0x00			; location of the queue which is written to the FDC
CMDRLOC		equ	0x10			; location of the bytes which are read from FDC

; Data buffer memory organization: starts at address 0x100 (addressed using indirect addressing scheme)

; System variables
CTRSTAT		equ	0x20			; controller status - free/busy
TMP			equ	0x21
MAINSTAT	equ	0x22			; last read master status register of FDD
SENSEST		equ	0x23			; last read drive sense status (status reg 3 - ST3)
FDCQPOS		equ	0x24
QTEMP		equ	0x25
CMDRUN		equ	0x26			; not zero if command is running
FDCRUNPOS	equ	0x27			; position of the pointer when writing sequence to the FDD controller
RESULTN		equ	0x28			; number of result bytes to read from FDD

FMTTMP1		equ	0x29
FMTTMP2		equ	0x2A
FMTTMP3		equ	0x2B

DEBUGN		equ	0x2C

TIMOUT1		equ	0x2D			; used within timout counting code
TIMOUT2		equ	0x2E			; used within timout counting code
FDCERR		equ	0x2F			; not 0 if error
DATAVAL		equ	0x30
MOTATO		equ	0x31			; motor drive A timeout (0=no timeout)
MOTBTO		equ	0x32			; motor drive B timeout (0=no timeout)
DORVAL		equ	0x33			; current value of the DOR
CURDRV		equ	0x34			; Current _physical_ disk drive (0=A, 1=B; always 0 for single drive config)

RECLENL		equ	0x35
RECLENH		equ	0x36

;DMADIR		equ	0x37			; direction for DMA transfer (0=read from FDC [default], 1=write to FDC)
HOSTCMD		equ	0x38
TEMPREQ		equ	0x39			; temporary value for forming request to FDC
TEMPRE1		equ	0x3A			; temporary value for forming request to FDC
COUVAL		equ	0x3B			; command/MC-host interface counter value
DRINUM		equ	0x3C			; number of physical drives connected to the FDC
DRVPRM		equ	0x3D			; _physical_ drive and head parameter in FDC format
CYLPRM		equ	0x3E			; cylinder parameter in FDC format
SECPRM		equ	0x3F			; sector parameter in FDC format
FMTPRM		equ	0x40
FMTPR1		equ	0x41			; three variables for format's writing control records
FMTPR2		equ	0x42
FMTPR3		equ	0x43
CURHOSTDRV	equ	0x44			; current _logical_ disk drive (0=A, 1=B)

SPDCOU		equ	0x45			; Counter for speed retry
DSK0SPD		equ	0x46			; DSK*SPD contain speed current disk/disk drive is using
DSK1SPD		equ	0x47			; 0=500kB/s, 1=250kB/s
CURGPL1		equ	0x48
CURGPL2		equ	0x49
CURDSP1		equ	0x4A
CURDSP2		equ	0x4B
CURDSPD		equ	0x4C
CURSPT		equ	0x4D
DSK0CHGFLG	equ	0x4E			; disk change flags for drives 0 and 1
DSK1CHGFLG	equ	0x4F

; DSKIO dedicated variables
DSKIO_SEC_0	equ	0x50
DSKIO_SEC_1	equ	0x51
DSKIO_SEC_2	equ	0x52
DSKIO_SEC_3	equ	0x53			; logical sector to access (4 bytes)
DSKIO_SpT	equ	0x54			; sectors per track (physical dimensions, 1 byte)
DSKIO_Heads	equ	0x55			; heads (physical dimensions, 1 byte)
DSKIO_NumOfSec	equ	0x56		; number of sectors to process
DIO_NumSecProcessed	equ	0x57	; number of sectors actually processed
DIO_Retry_Count	equ	0x58		; retry count for the operation
CURCYL0		equ	0x59
CURCYL1		equ	0x5A			; current cylinder (head) positions of _physical_ drives
DSKIO_CRC	equ	0x5b


TIMOUT3		equ	0x5c

DELAY1		equ	0x5D	; used within delay counting code
DELAY2		equ	0x5E	; used within delay counting code
DELAY3		equ	0x5F	; used within delay counting code
; --- Last variable space offset within ACCESS bank = 0x5F


; PROGRAM START
	org	0
	goto	main

; INTERRUPT VECTOR SETTING
	org	8
	goto	int_handler

Set144SpeedVals:
	movlw	GPL1_144
	movwf	CURGPL1,ACCESS
	movlw	GPL2_144
	movwf	CURGPL2,ACCESS
	movlw	DSPEC1_144
	movwf	CURDSP1,ACCESS
	movlw	DSPEC2_144
	movwf	CURDSP2,ACCESS
	movlw	DSPD_144
	movwf	CURDSPD,ACCESS
	movlw	SpT_144
	movwf	CURSPT,ACCESS
	return

Set720SpeedVals:
	movlw	GPL1_720
	movwf	CURGPL1,ACCESS
	movlw	GPL2_720
	movwf	CURGPL2,ACCESS
	movlw	DSPEC1_720
	movwf	CURDSP1,ACCESS
	movlw	DSPEC2_720
	movwf	CURDSP2,ACCESS
	movlw	DSPD_720
	movwf	CURDSPD,ACCESS
	movlw	SpT_720
	movwf	CURSPT,ACCESS
	return

; ACCESS space: bank 0 addresses 0..5F (0..5F), bank 15 addresses 60..FF (F60..FFF)
; important: ANSELx locations are out of the access space (below abslote address F60)!

main:
	movlw	1
	movwf	DSK0CHGFLG,ACCESS		; disk changed on startup
	movwf	DSK1CHGFLG,ACCESS

	clrf	DSK0SPD,ACCESS
	clrf	DSK1SPD,ACCESS			; setting speed 500kB/s

	; setting default speeds
	call	Set720SpeedVals			; default configuration: Legacy 720

	clrf	DEBUGN,ACCESS

	; set timter TMR1
	clrf	TMR1H,ACCESS
	clrf	TMR1L,ACCESS

	; turning timer1 on (driven by internal clock fosc/4)
	movlw	0x11					; timer 1 on, prescaler 1:2 (slower interrupt rate and thus
	movwf	T1CON,ACCESS			; longer time for drive motor to shut down
	movlw	0x01
	movwf	PIE1,ACCESS				; timer1 overflow interrupt enable
	movlw	0xC0
	movwf	INTCON,ACCESS			; enable interrupts globall

	; increasing running speed to 64MHz
	movlw	0xf0					; internal oscillator, 16MHz
	movwf	OSCCON,ACCESS			; PLL on, 32MHz
	movlw	0x5f
	movwf	OSCTUNE,ACCESS			; maximal frequency with PLL enabled (CONFIG1H sets INTIO67)

	; all ditigal pins setup
	movlb	0x0f
	clrf	ANSELA,BANKED
	clrf	ANSELB,BANKED
	clrf	ANSELC,BANKED
	clrf	ANSELD,BANKED
	clrf	ANSELE,BANKED			; all ports are digital outputs and inputs
	movlb	0

	movlw	0xFD					; RE1 as output, other pins as input
	movwf	TRISE,ACCESS
	setf	TRISD,ACCESS			; port D as input
	clrf	TRISC,ACCESS
	clrf	TRISB,ACCESS
	clrf	TRISA,ACCESS			; ports A, B, C as outputs

	setf	PORTA,ACCESS
	setf	PORTB,ACCESS
	setf	PORTC,ACCESS
	setf	PORTD,ACCESS
	setf	PORTE,ACCESS			; all ports A-D output high level (LATAx=0xFF)

	; clear controller status flag
	clrf	CTRSTAT,ACCESS
	clrf	CMDRUN,ACCESS
	clrf	MOTATO,ACCESS
	clrf	MOTBTO,ACCESS
	clrf	DRINUM,ACCESS
	clrf	CURDRV,ACCESS
	clrf	CURHOSTDRV,ACCESS

LoopReset:
	movf	PORTE,W,ACCESS
	andlw	0xC					; RE3/RE2
	btfss	STATUS,Z,ACCESS
	goto	LoopReset			; loop until both are 0 (inactive)

	movlw	0xFF
	call	Wait_ms				; wait 255 ms

HostWaitMainLoop:
	bsf		INTCON,GIE,ACCESS		; enable interrupts
	call	WaitForHostWrite		; wait for host write. Interrupts are enabled, thus host write may be missed by the MC.
									; If MC misses command write, then host will next read 0xFF. If MC did not miss command write
									; it disables interrupts and returns command initiation byte 0xAA (or 0xAB, depending
									; on request)

	; process looks the following way:
	; 1. host ensures that MC is available/free through using loopback command
	; 2. host issues write command with command needed to port base+0. If MC was in the process of some command, it aborts
	;    the command and returns to HostWaitMainLoop
	; 3. host writes parameters to the port base+1 of MC. Host can write anything to port base+0 and this action aborts the sequence
	;    returning 0xAA and going to HostWaitMainLoop
	; 4. As soon as MC receives required set of parameters, it starts processing request. Depending on command MC returns command execution
	;    byte(s) to the host. Host checks progress by reading from base port+1.
	; 5. When required number of bytes were transferred, host starts to receive status information.
	; 6. When all return data is read, MC returns to HostWaitMainLoop and waits for next command. Interrupts are then enabled, and disk drive motor
	;    delay countdown is processed in the interrupt routine.

	; examination of the requested operation (WREG)
	; 0xAA = reset command progress and echo/looback (returns 0xAA to the host)
	; 0x00 = INIFDD, initialize FDC and FDDs (confirmation parameter 0xEC)
	; 0x01 = DSKIO read sector (drive#+head, cylinder, sector number; returns MC standard error data, see below)
	; 0x02 = DSKIO write sector (same as read sector)
	; 0x03 = DSKCHG check for disk change (drive#; returns 0xFF disk changed, 0x01 disk unchanged)
	; 0x04 = DSKFMT disk format (confirmation parameter 0xEC, and then drive#)
	; 0x05 = DRIVES return number of physical drives and currently selected drive (0-3)
	; 0x06 = VERSION returns MC version string
	; 0x07 = SOFTWARE read software package (MSXDOS,MSXDOS2, COMMAND and COMMAND2)
	; 0x08 = MAKEDPB return DPB structure constructed out of data in buffer at 0x100 (boot sector should be in there!)
	; 0x09 = HDDSTATUS return selected HDD status register
	; 0x0a = Unimplemented - HDDCHS read hard disk's sector in CHS mode
	; 0x0b = Unimplemented - HDDCHS write hard disk's sector in CHS mode
	; 0x0c = HDDRST
	; 0x0d
	; 0x0e
	; 0x0f
	; 0x10 = HDDID hard disk initialize and identify command (returns 512 bytes)
	; 0x11 = BUFACC read contents of the data buffer
	; 0x12 = BUFACC write to the data buffer
	; > 0x0a = reset request, return 0xAA to the host and restart sensing new request

	btfsc	STATUS,C,ACCESS			; A0 address line state during host request
	goto	HostWaitMainLoop		; access to port base+1 is not assumed this time

	bcf		INTCON,GIE,ACCESS		; disable interrupts

	movwf	HOSTCMD,ACCESS
	sublw	0xAA
	btfsc	STATUS,Z,ACCESS
	goto	NoCommandToRun

	movf	HOSTCMD,W,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	INIFDD					; 0x00
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKIO					; 0x01
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKIO					; 0x02
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKCHG					; 0x03
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DSKFMT					; 0x04
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	DRIVES					; 0x05
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	VERSION					; 0x06
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	SOFTWARE				; 0x07
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	MAKEDPB					; 0x08
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	HDDSTAT					; 0x09

	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	NoCommandToRun			; 0x0a
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	NoCommandToRun			; 0x0b
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	HDDRST			; 0x0c
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	NoCommandToRun			; 0x0d
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	NoCommandToRun			; 0x0e
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	NoCommandToRun			; 0x0f

	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	HDDID					; 0x10
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	BUFACC					; 0x11
	decf	WREG,1,ACCESS
	btfsc	STATUS,Z,ACCESS
	goto	BUFACC					; 0x12

NoCommandToRun:
	movlw	0xAA
	call	WaitForHostRead				; returns 0xAA to the host = "loopback", or "no command to run"
	bra		HostWaitMainLoop

; --- Read or write contents of the MC data buffer
BUFACC:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	call	WaitForHostWrite
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; number of bytes to transfer (LOW), use FSR2 as 16-bit counter
	movwf	FSR2L,ACCESS

	call	WaitForHostWrite
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; number of bytes to transfer (HIGH), use FSR2 as 16-bit counter
	movwf	FSR2H,ACCESS

	; check for byte count to be 0
	iorwf	FSR2L,W,ACCESS
	bz		BUF_Ret_Error				; return 0xAF with result code 0 immediately (further two writes of
										; buffer start address are automatically discarded)
	call	WaitForHostWrite
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; buffer start address (LOW) - important: buffer for host starts at 0x0000 (within MC it is at 0x0100)
	movwf	FSR1L,ACCESS

	call	WaitForHostWrite
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; buffer start address (HIGH) - important: buffer for host starts at 0x0000 (within MC it is at 0x0100)
	incf	WREG,W,ACCESS
	movwf	FSR1H,ACCESS

	; signalling to host that MC is ready to transfer data
	movlw	0xAE
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	; *** clear CRC byte
	clrf	DSKIO_CRC,ACCESS

	; *** checking if read or write was requested
	btfsc	HOSTCMD,1,ACCESS			; bit 1: set = write to MC, reset = read from MC
	bra		BUFWdata

BUFRdata:
	movf	POSTINC1,W,ACCESS
	addwf	DSKIO_CRC,1,ACCESS				; add data byte to CEC
	call	WaitForHostRead				; send byte to host

	movf	POSTDEC2,W,ACCESS			; decrement 16-bit pair FSR2H:FSR2L
	movf	FSR2L,W,ACCESS				; get FSR2L into WREG
	iorwf	FSR2H,W,ACCESS				; OR with FSR2H
	bnz		BUFRdata					; if not zero - continue looping
	bra		BUFFinish					; otherwise - exit

BUFWdata:
	call	WaitForHostWrite			; read byte from host
	addwf	DSKIO_CRC,1,ACCESS				; add data byte to CEC
	movwf	POSTINC1,ACCESS

	movf	POSTDEC2,W,ACCESS			; decrement 16-bit pair FSR2H:FSR2L
	movf	FSR2L,W,ACCESS				; get FSR2L into WREG
	iorwf	FSR2H,W,ACCESS				; OR with FSR2H
	bnz		BUFWdata					; if not zero - continue looping
	bra		BUFFinish					; otherwise - exit

BUFFinish:
	; now checking data CRC
	call	WaitForHostWrite		; read CRC from host
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	return							; access to port base+1 is not assumed this time, aborting MC command

	subwf	DSKIO_CRC,1,ACCESS
	movlw	0xCC
	btfss	STATUS,Z,ACCESS
BUF_Ret_Error:
	movlw	0x00					; CRC does not match: error

	movwf	SECPRM,ACCESS				; restore return byte

	movlw	0xAF
	call	WaitForHostRead				; returning 0xAF - result phase
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	SECPRM,W,ACCESS				; restore return byte
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop


; --- Constructs DPB from the data in the MC buffer (starting address 0x100),
; this means that host should first successfully read boot sector into the buffer
; important: calling this procedure will invalidate portion of data starting 0x80 absolute sector offset
; located at 0x100
MAKEDPB:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movlw	0xcc
	movwf	SPDCOU						; output value - if not modified troughout execution

	; setting up pointer to the output buffer
	movlw	0x1
	movwf	FSR0H,ACCESS
	movlw	0x80
	movwf	FSR0L,ACCESS				; point to the temp location 0x180

	movlb	0x01						; bank 1
	; move media desctiptor
	movf	0x15,W,BANKED
	movwf	POSTINC0,ACCESS				; +1: media descriptor
	; check sector size
	movf	0x0b,W,BANKED
	bnz		ConfErr
	movwf	POSTINC0,ACCESS				; +2 SECSIZ LOW of the sector size
	movlw	0x02
	cpfseq	0x0c,BANKED
	bra		ConfErr
	movwf	POSTINC0,ACCESS				; +3 SECSIZ HIGH of the sector size
	movlw	0x0f
	movwf	POSTINC0,ACCESS				; +4 DIRMSK=(SECSIZ/32)-1 directory mask
	movlw	0x04
	movwf	POSTINC0,ACCESS				; +5 DIRSHFT mask bits for DIRMSK
	movf	0x0d,W,BANKED
	bz		ConfErr						; sectors per cluster can not be 0
	decf	WREG,W,ACCESS
	movwf	POSTINC0,ACCESS				; +6 CLUSMSK = (Sectors per cluster)-1
	call	CalcMaskForByte
	movwf	POSTINC0,ACCESS				; +7 CLUSSHFT = (Number of one bits in CLUSMSK)+1
	movf	0x0e,W,BANKED
	movwf	POSTINC0,ACCESS
	movf	0x0f,W,BANKED
	movwf	POSTINC0,ACCESS				; +8 and +9: FIRFAT = Logical sector number of first FAT
										; equals to number of reserved sectors
	movlw	0x03
	cpfslt	0x10,BANKED
	bra		ConfErr						; number of FAT copies can not be more than 2
	movf	0x10,W,BANKED
	bnz		NoConfErr					; number of FAT copies can not be 0
ConfErr:
	clrf	SPDCOU,ACCESS				; return code = 0x00 (total error)
	bra		MAKEDPB_return

NoConfErr:
	movwf	POSTINC0,ACCESS				; +10 FATCNT = Number of FATs (supported values: 1 and 2)

	mulwf	0x16,BANKED					; (number of sector in FAT) * (FAT count)
	movf	0x0e,W,BANKED
	addwf	PRODL,1,ACCESS
	movf	0x0f,W,BANKED
	addwfc	PRODH,1,ACCESS				; PROD = (number of sector in FAT) * (FAT count) + FAT starting sector
										; e.g. points to the start of directory
	movf	PRODL,W,ACCESS
	movwf	0x90,BANKED
	movf	PRODH,W,ACCESS
	movwf	0x91,BANKED					; directly copy starting sector of directory
										; to the DPB being formed (without using INDF0)
	movf	0x12,1,BANKED
	bnz		TruncDir					; number of directory entries can not be more than 256, go truncate
	movf	0x11,W,BANKED
	bz		ConfErr						; number of directory entries can not be 0 - error
	incf	WREG,W,ACCESS
	bz		TruncDir					; number of dir entries = 0xff - go truncate to 254
	decf	WREG,W,ACCESS
	bra		DirNot255

TruncDir:
	; directory entries are more than 254 - truncating
	movlw	0xcf
	movwf	SPDCOU,ACCESS				; return code = 0xcf - partial error (dir truncated)
	movlw	0xfe						; maximal number of entries - 254

DirNot255:
	movwf	POSTINC0,ACCESS				; +11: MAXENT = Number of directory entries (max 254)

	; now calculate number of sectors in DIR (real, not truncated!)
	movf	0x12,W,BANKED
	andlw	0x0f
	swapf	WREG,W,ACCESS
	movwf	TMP,ACCESS					; TMP = byte HI low nibble high
	movf	0x11,W,BANKED
	swapf	WREG,W,ACCESS
	andlw	0x0f
	iorwf	TMP,W,ACCESS				; it is directory entries word shifted
										; right 4 bits (i.e. /16 = sectors in DIR)
	addwf	PRODL,1,ACCESS
	clrf	WREG,ACCESS					; instruction does not affect C flag
	addwfc	PRODH,W,ACCESS				; W:PRODL = (number of sector in FAT) * (FAT count) + FAT starting sector + number of DIR sectors
	movff	PRODL,POSTINC0
	movwf	POSTINC0,ACCESS				; +12 and +13: FIRREC = Logical sector number of first data

	; now calculate number of clusters
	movf	PRODL,W,ACCESS
	subwf	0x13,W,BANKED
	movwf	FMTPRM,ACCESS
	movf	PRODH,W,ACCESS
	subwfb	0x14,W,BANKED
	movwf	FMTPR1,ACCESS				; number of sectors on the disk minus number of control sectors (boot, FATs, dir)
	movf	0x0d,W,BANKED				; sectors per cluster
	movlb	0xf
	movwf	DVSR_0,BANKED
	clrf	DVSR_1,BANKED				; divisor
	movff	FMTPRM,DVDT_0
	movff	FMTPR1,DVDT_1				; dividend
	call	DIV16						; data sectors / sectors per cluster
	incf	DVDT_0,W,BANKED
	movwf	POSTINC0,ACCESS
	clrf	WREG,ACCESS
	addwfc	DVDT_1,W,BANKED
	movwf	POSTINC0,ACCESS				; +14 and +15: MAXCLUS = maximal number of cluster
	movlb	0x01

	movf	0x17,1,BANKED
	bnz		ConfErr						; sectors per FAT can not be more than 256
	movf	0x16,W,BANKED
	bz		ConfErr						; sectors per FAT can not be 0
	sublw	0x0a
	bnc		ConfErr						; sectors per FAT can not be more than 9
	movf	0x16,W,BANKED
	movwf	POSTINC0,ACCESS				; +16: FATSIZ = number of sectors used for FAT

MAKEDPB_return:
	movlw	0xAF
	call	WaitForHostRead				; returning 0xAF - result phase
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	SPDCOU,W,ACCESS				; restore return byte
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop


; calculates mask (how many effective bits are in the representation of the number)
; in:	WREG=value to calculatye effective bits for
; out:	WREG=number of bits+1
CalcMaskForByte:
	movwf	COUVAL,ACCESS				; store input value
	movlw	0x08						; 8 bits to process
CalcBitLook:
	rlcf	COUVAL,1,ACCESS				; rotate input value left through carry
										; store value back
	bc		CalcBitFound
	decfsz	WREG,W,ACCESS
	bra		CalcBitLook
CalcBitFound:
	incf	WREG,W,ACCESS
	return


; --- Return software image
SOFTWARE:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	call	WaitForHostWrite			; wait for the package ID to return
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movwf	FMTPRM,ACCESS				; paramater is: bits 7 = COMMAND (1), MSXDOS (0); bit 6= DOS2 (1), DOS1 (0)
	movwf	FMTPR1,ACCESS				; bits 5..0 (total 64) - number of 512 byte page of the image (thus max image = 32K)

	rlncf	FMTPRM,1,ACCESS				; page*2
	movlw	0x7E
	andwf	FMTPRM,1,ACCESS				; reset bits 0 and 7, now FMTPRM contains addition to TBLPTRH

	movf	FMTPR1,W,ACCESS				; restore original SW ID parameter
	rlncf	WREG,1,ACCESS
	rlncf	WREG,1,ACCESS				; now image selection ID is in bits 0 and 1
	andlw	0x03						; mask two LSBs
	bnz		NotMSXDOSSYS

	; image bits 00 = MSXDOS.SYS
	movlw	LOW MSXDOS
	movwf	TBLPTRL,ACCESS
	movlw	HIGH MSXDOS
	addwf	FMTPRM,W,ACCESS				; need to keep CY to add to Upper part of counter
	movwf	TBLPTRH,ACCESS
	movlw	UPPER MSXDOS
	btfsc	STATUS,C
	addlw	1
	movwf	TBLPTRU,ACCESS
	bra		SendSector

NotMSXDOSSYS:
	decfsz	WREG,1,ACCESS
	bra		NotMSXDOS2SYS

	; image bits 01 = MSXDOS2.SYS
	movlw	LOW MSXDOS2
	movwf	TBLPTRL,ACCESS
	movlw	HIGH MSXDOS2
	addwf	FMTPRM,W,ACCESS				; need to keep CY to add to Upper part of counter
	movwf	TBLPTRH,ACCESS
	movlw	UPPER MSXDOS2
	btfsc	STATUS,C
	addlw	1
	movwf	TBLPTRU,ACCESS
	bra		SendSector

NotMSXDOS2SYS:
	decfsz	WREG,1,ACCESS
	bra		NotCOMMANDCOM

	; image bits 10 = COMMAND.COM
	movlw	LOW COMMAND
	movwf	TBLPTRL,ACCESS
	movlw	HIGH COMMAND
	addwf	FMTPRM,W,ACCESS				; need to keep CY to add to Upper part of counter
	movwf	TBLPTRH,ACCESS
	movlw	UPPER COMMAND
	btfsc	STATUS,C
	addlw	1
	movwf	TBLPTRU,ACCESS
	bra		SendSector

NotCOMMANDCOM:
	; image bits 11 = COMMAND2.COM
	movlw	LOW COMMAND2
	movwf	TBLPTRL,ACCESS
	movlw	HIGH COMMAND2
	addwf	FMTPRM,W,ACCESS				; need to keep CY to add to Upper part of counter
	movwf	TBLPTRH,ACCESS
	movlw	UPPER COMMAND2
	btfsc	STATUS,C
	addlw	1
	movwf	TBLPTRU,ACCESS

SendSector:
	movlw	0xAC
	call	WaitForHostRead				; return 0xAC to the host - ready to transfer 

	clrf	FMTPRM,ACCESS				; CRC
	clrf	RECLENL,ACCESS
	movlw	0x02
	movwf	RECLENH,ACCESS

SWTransferLoop:
	tblrd*+
	movf	TABLAT,W,ACCESS
	addwf	FMTPRM,1,ACCESS
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	HostWaitMainLoop			; access to port base+1 is not assumed this time, aborting MC command

	decfsz	RECLENL,1,ACCESS
	bra		SWTransferLoop
	decfsz	RECLENH,1,ACCESS			; increase byte count
	bra		SWTransferLoop

	call	WaitForHostWrite			; wait for hosts's CRC
	subwf	FMTPRM,W,ACCESS
	movlw	0xCC						; success
	btfss	STATUS,Z,ACCESS
	movlw	0x00						; error
	call	WaitForHostRead				; return status code
	goto	HostWaitMainLoop

; --- Give own version to the host
VERSION:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	; and now return ID string hardcoded in MC's EPROM
	movlw	UPPER IDTable
	movwf	TBLPTRU,ACCESS
	movlw	HIGH IDTable
	movwf	TBLPTRH,ACCESS
	movlw	LOW IDTable
	movwf	TBLPTRL,ACCESS

IDLoop:
	tblrd*+
	movf	TABLAT,W,ACCESS
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	HostWaitMainLoop			; access to port base+1 is not assumed this time, aborting MC command

	incf	TABLAT,W,ACCESS				; 0xff + 1 = 0 raises STATUS's bit Z
	bnz		IDLoop
	goto	HostWaitMainLoop

IDTable:
	db		"PIC18F46K22 v.2.0a",0xff

; --- INIFDD host routine
INIFDD:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	call	WaitForHostWrite			; read confirmation code
	btfss	STATUS,C,ACCESS				; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun				; in this case return 0xAA

	sublw	0xEC						; pass word to start initialization of the FDD hardware
	btfss	STATUS,Z,ACCESS
	goto	NoCommandToRun

	; reset disk change flags (this INIFDD command is usually
	; invoked during system initialization)
	movlw	1
	movwf	DSK0CHGFLG,ACCESS		; disk changed on startup
	movwf	DSK1CHGFLG,ACCESS

	; *** initialize FDD subsystem
	; wait until system reset condition is cleared
	call	SoftReset

	clrf	DRINUM,ACCESS
	clrf	CURDRV,ACCESS
	call	InitDrive			; init drive 0 (A)
	btfss	STATUS,C,ACCESS
	incf	DRINUM,1,ACCESS		; now drive count = 1
	incf	CURDRV,1,ACCESS
	call	InitDrive			; init drive 1 (B)
	btfss	STATUS,C,ACCESS
	incf	DRINUM,1,ACCESS		; now drive count = 2

	clrf	CURDRV,ACCESS		; by default after hardware init drive 0 is selected
	clrf	CURHOSTDRV,ACCESS

	movlw	0x80
	call	WaitForHostRead				; return command completion code to the host
	goto	HostWaitMainLoop

; --- DRIVES host routine
DRIVES:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	DRINUM,W,ACCESS
	call	WaitForHostRead				; return number of drives to host
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	CURHOSTDRV,W,ACCESS
	call	WaitForHostRead				; return current drive selected (used for single physical drive config)
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop

; --- DSKIO host routine
; returns: 0x0C (12 = unknown error) if disk was changed, 0x02 (not ready) in case of MC timeout, 0x04 (crc error)
;			on data transfer CRC error
;			other error codes according to the FDC bits set
DSKIO:
	movlw	0xAB
	call	WaitForHostRead				; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	; *** getting parameters for command
	call	WaitForHostWrite			; get first parameter: drive # and bit 6 as device type ID (0=FDD, 1=HDD)
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; drive parameter: bit 6 (0:FDD, 1:HDD), bit 0 (0: fdd0/hdd0, 1: fdd1/hdd1)
	movwf	DRVPRM,ACCESS

	call	WaitForHostWrite			; get second parameter: cylinder#
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; absolute sector byte 0 (LOW)
	movwf	DSKIO_SEC_0,ACCESS

	call	WaitForHostWrite			; get second parameter: cylinder#
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; absolute sector byte 1
	movwf	DSKIO_SEC_1,ACCESS

	call	WaitForHostWrite			; get second parameter: cylinder#
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; absolute sector byte 2
	movwf	DSKIO_SEC_2,ACCESS

	call	WaitForHostWrite			; get second parameter: cylinder#
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; absolute sector byte 3 (HIGH)
	movwf	DSKIO_SEC_3,ACCESS

	call	WaitForHostWrite			; get second parameter: cylinder#
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; Sectors per track
	movwf	DSKIO_SpT,ACCESS

	call	WaitForHostWrite			; get second parameter: cylinder#
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	; Number of heads
	movwf	DSKIO_Heads,ACCESS

	call	WaitForHostWrite			; get third parameter: sector# (track starts with sector 0x01)
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	movwf	DSKIO_NumOfSec,ACCESS
	movlw	0x07						; max 7 sectors, otherwise MC is dead (overwrite of control registers)
	cpfslt	DSKIO_NumOfSec,ACCESS
	movwf	DSKIO_NumOfSec,ACCESS		; =7 if requested to process more than 7 sectors!

	; *** initialize number of sectors processed
	clrf	DIO_NumSecProcessed,ACCESS

	; *** all parameters received, start working...
	movf	DRVPRM,W,ACCESS
	btfsc	WREG,6,ACCESS				; checking hard disk bit set?
	bra		DIO_HDD_Operation

	; *** we are here if FDD operation is requiested (bit 6 is 0)
	andlw	0x01
	call	CheckSingleDrive		; updates CURHOSTDRV (_logical_ drive number), and forces
									; CURDRV (_physical_ dribe number) to be 0 for single drive configuration
	call	MotorOnSelect			; turn _physical_ drive motor on

	; - currently DRVPRM contains logical drive number (0 or 1)
	; - and CURDRV contains physical drive number (always 0 for single drive config,
	; and 0 or 1 for 2 drive config)
	; - CURHOSTDRV is a global system variable containing logical drive selected by host

	; *** check if write operation was requested, and diskette is write protected
	movf	HOSTCMD,W,ACCESS
	decf	WREG,W,ACCESS
	btfsc	STATUS,Z,ACCESS
	bra		DoNotCheckWriteProtected

	; *** write operation - check for disk write protection issuing SENSE DRIVE STATUS command
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS			; checking state of the physical drive
	call	AddToFDCQueue
	call	RunCommand

	btfsc	FDCERR,0,ACCESS			; timeout - "not ready" error
	bra		DSKIO_ERR_NotReady

	; ST3:
	; bit 6=0: not write protected
	; bit 6=1: write protected
	btfss	CMDRLOC,6,ACCESS		; if bit 6 of ST3=1 then disk is write protected
	bra		DoNotCheckWriteProtected
	
	; otherwise diskette is write protected
	clrf	WREG,ACCESS				; WRITE PROTECT error
	bra		DSKIO_ReturnStatus		; get out completely, no retries

DoNotCheckWriteProtected:
	; *** set buffer pointer to the beginning of data space: uses FSR1
	clrf	FSR1L,ACCESS
	movlw	0x01
	movwf	FSR1H,ACCESS
DIO_NextFullLoop:
	; *** initialize retry count
	movlw	RETRIES
	movwf	DIO_Retry_Count,ACCESS
Next_Retry:
	; *** initialize speed count loop
	movlw	2
	movwf	SPDCOU,ACCESS			; trying two speeds

SpeedTryLoop:
	; *** set controller data transfer speed
	movf	DSK0SPD,W,ACCESS
	btfsc	CURDRV,0,ACCESS			; CURDRV (physical drive)
									; (which is ID of physical drive and is 0 for phantom drive)
	movf	DSK1SPD,W,ACCESS

	btfss	WREG,0,ACCESS			; check remembered value for speed for specific drive (0/1)
	bra		Set144					; bit 0=0 is for 500kb/s, 1 for 250kb/s
	call	Set720SpeedVals
	bra		SpeedsSet
Set144:
	call	Set144SpeedVals

SpeedsSet:
	call	SetSpeeds				; set speeds for specific drive according to its remembered mode

	; *** now check for disk changed, and for need to recalibrate
	; after previous unsuccessful attempt to read or write
	call	ReadDiskChange
	btfsc	TEMPREQ,7,ACCESS
	bra		DskChanged				; disk change bit set - proceed with seek and see if it will be still set

	movlw	RETRIES
	cpfslt	DIO_Retry_Count,ACCESS
	bra		NotChanged				; retry counter is max - it is first time operation occurs,
									; and recalibration should not be performed
DskChanged:
	movlw	0x01
	movwf	CYLPRM,ACCESS
	call	SeekCommand				; Cylinder 1
	btfsc	FDCERR,0,ACCESS
	bra		DSKIO_ERR_NotReady			; recalibration error

	call	Recalibrate				; recalibrate
	btfsc	FDCERR,0,ACCESS
	bra		DSKIO_ERR_NotReady		; recalibration error

	call	ReadDiskChange
	btfss	TEMPREQ,7,ACCESS
	bra		NotChanged

	movlw	0x0C
	bra		DSKIO_ReturnStatus		; return 0x0C and exit completely, with no retries
									; (as it is indicator for missing diskette in drive)
NotChanged:
	; *** saving buffer pointer FSR1 to FSR2 - in case it will be changed during
	; unsuccessful command execution we will need to revert it back to original position
	movff	FSR1L,FSR2L
	movff	FSR1H,FSR2H

	; *** presetting _physical_ drive number for read/write command
	movf	CURDRV,W,ACCESS
	movwf	DRVPRM,ACCESS

	; *** now we will calculate C/H/S to send to drive, using 16-bit math
	; first, calculate disk's SpT*Heads
	movf	DSKIO_SpT,W,ACCESS		; get number of sectors per track
	btfss	DSKIO_Heads,1,ACCESS	; checking bit 1 of the number of heads
	bra		DSKIO_OnlyOneHead

	addwf	WREG,W,ACCESS			; SpT*2, note limitation: SpT can not be more than 127

DSKIO_OnlyOneHead:
	; now calculate cylinder Cyl=LS/(Sec*Hd)
	movlb	0x0f					; bank 15
	movff	DSKIO_SEC_0,DVDT_0
	movff	DSKIO_SEC_1,DVDT_1
	movff	WREG,DVSR_0				; (Sec*Hd) => divisor
	clrf	DVSR_1,BANKED
	call	DIV16					; cylinder number will be in DVDT (expect 8-bit value)
	movf	DVDT_0,W,BANKED
	movwf	CYLPRM,ACCESS			; now we have cylinder humber
	; now calculate head Hd=(LS%(Sec*Hd))/Sec
	movff	RMDR_0,DVDT_0			; assume remainder was also 8-bit
	clrf	DVDT_1,BANKED
	movff	DSKIO_SpT,DVSR_0
	clrf	DVSR_1,BANKED
	call	DIV16
	; process head number
	movf	DVDT_0,W,BANKED
	rlncf	WREG,W,ACCESS
	rlncf	WREG,W,ACCESS
	andlw	0x04					; mask head bit
	iorwf	DRVPRM,1,ACCESS			; and set head bit in the parameter for FDC read/write command
	; process sector number
	incf	RMDR_0,W,BANKED			; remainder +1 (FDC sector number starts with 1)
	movwf	SECPRM,ACCESS
	movlb	0x00					; return back to bank 0

	; *** performing seek (checkinf if head is already there beforehand)
	call	GetPhyDriveCurCyl
	subwf	CYLPRM,W,ACCESS
	bz		DIO_AlreadyAtCylinder	; skip seek command if heads are already at the required cylinder
									; this will save some operating time...
	call	SeekCommand
	btfsc	FDCERR,0,ACCESS
	bra		DSKIO_ERR_Seek

DIO_AlreadyAtCylinder:
	; *** and finally issue command to the FDC
	call	NewFDCRequest			; clear MC FDC queue
	movf	HOSTCMD,W,ACCESS
	decf	WREG,W,ACCESS
	movlw	0x46					; HOSTSCMD=1 -- read operation 0x46
	btfss	STATUS,Z,ACCESS
	movlw	0x45					; HOSTSCMD=2 -- write operation 0x45
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS			; head # + drive # (in FDC format)
	call	AddToFDCQueue
	movf	CYLPRM,W,ACCESS			; cylinder #
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS
	rrncf	WREG,1,ACCESS
	rrncf	WREG,1,ACCESS
	andlw	1						; head #
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS			; sector #
	call	AddToFDCQueue
	movlw	0x02					; number of bytes (0x200 = 512)
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS			; last sector# on the track to transfer (the same as sector# to read/write)
	call	AddToFDCQueue
	movf	CURGPL1,W,ACCESS		; GAP1 default size
	call	AddToFDCQueue
	movlw	0xFF					; Data length (default 0xFF)
	call	AddToFDCQueue
	call	RunCommandWithData

	btfss	FDCERR,0,ACCESS
	bra		DIO_NoFDCRWError		; go there in case FDCERR is not set (no errors)

	movlw	0x08					; "Record not found" error for further processing

DIO_NextSpeed:
	; *** error occured - revert previous value of buffer pointer and try again
	movff	FSR2L,FSR1L
	movff	FSR2H,FSR1H

	; *** error occured, need to try another speed
	dcfsnz	SPDCOU,1,ACCESS
	bra		DIO_RetryCount			; go for next retry

	movlw	0x01
	btfss	CURDRV,0,ACCESS
	bra		ChgDrv0
	; drive 1
	xorwf	DSK1SPD,1,ACCESS
	bra		SpeedTryLoop
ChgDrv0:
	; drive 0
	xorwf	DSK0SPD,1,ACCESS
	bra		SpeedTryLoop

DIO_RetryCount:
	decfsz	DIO_Retry_Count,1,ACCESS
	bra		Next_Retry
	bra		DSKIO_ReturnStatus

DIO_NoFDCRWError:
	; *** continue if no R/W error occured
	call	DecodeError				; returns 0xCC if no error
	btfss	WREG,7,ACCESS			; bit 7 = 1 - no error occured
	bra		DIO_NextSpeed			; error occured: try another speed setting

DSKIOSuccRetStat:
	; *** we are here if successful operation on the sector accomplished
	; *** save the speed of the channel success was found with
	clrf	TMP,ACCESS				; speed ID=0 (500Kbit/s)
	movlw	SpT_144
	cpfseq	CURSPT,ACCESS			; is current sector per track = 1.44M config?
	incf	TMP,1,ACCESS			; now speed ID=1 (250Kbit/s)

	; and save this speed to respective physical disk drive config
	movf	TMP,W,ACCESS			; restore actual speed ID to save
	btfss	CURDRV,0,ACCESS
	bra		SavDrv0
	; drive 1
	movwf	DSK1SPD,ACCESS
	bra		DSKIOExitSuccess
SavDrv0:
	; drive 0
	movwf	DSK0SPD,ACCESS

DSKIOExitSuccess:
	; *** check if there're more sectors to read/write
	incf	DIO_NumSecProcessed,1,ACCESS
	movf	DIO_NumSecProcessed,W,ACCESS
	subwf	DSKIO_NumOfSec,W,ACCESS
	bz		DIO_AllSectorsProcessed

	; *** no, not all sectors were processed
	incf	DSKIO_SEC_0,1,ACCESS
	clrf	WREG,ACCESS
	addwfc	DSKIO_SEC_1,1,ACCESS	; next logical sector
	bra		DIO_NextFullLoop		; go back with logical sector +1, sector count-1,
									; and FSR1 shifted by 0200h
DSKIO_ERR_NotReady:
	movlw	0x02
	bra		DIO_NextSpeed			; go there trying another speed,
									; if exhausted - return 0x02 to host
DSKIO_ERR_Seek:
	movlw	0x06
	bra		DIO_NextSpeed			; go there trying another speed,
									; if exhausted - return 0x06 to host
DIO_AllSectorsProcessed:
	movlw	0xCC

DSKIO_ReturnStatus:
	; *** return control data and exit
	call	MotorOff

HDD_ReturnStatus:
	movwf	SECPRM,ACCESS			; preserve return byte
	movlw	0xAF
	call	WaitForHostRead			; returning 0xAF - result phase
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	DIO_NumSecProcessed,W,ACCESS

	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	SECPRM,W,ACCESS				; restore return byte
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop

FlashN:
	movwf	RECLENH,ACCESS
	movlw	0x4
	movwf	RECLENL,ACCESS
FlashNCycle:
	movff	RECLENH,PORTB
	movlw	0xff
	call	Wait_ms
	movlw	0xff
	call	Wait_ms
	clrf	PORTB,ACCESS
	movlw	0xff
	call	Wait_ms
	movlw	0xff
	call	Wait_ms
	decfsz	RECLENL,1,ACCESS
	bra		FlashNCycle
	movf	RECLENH,W,ACCESS
	return

DIO_HDD_Operation:
	andlw	0x01					; reset HDD bit
	swapf	WREG,W,ACCESS
	movwf	DRVPRM,ACCESS			; drive number in HDD format (bit 4)

	; reset buffer pointer
	movlw	0x01
	movwf	FSR0H,ACCESS	
	clrf	FSR0L,ACCESS			; buffer start at 0x100

DIO_HDD_NextFullLoop:
	; *** now we will calculate C/H/S to send to drive, using _32_-bit math
	; first, calculate disk's SpT*Heads (16 effective bits)
	movf	DSKIO_SpT,W,ACCESS
	mulwf	DSKIO_Heads,ACCESS		; result is in PRODH:PRODL

	; now calculate cylinder Cyl=LS/(Sec*Hd)
	movlb	0x0f					; bank 15
	movff	DSKIO_SEC_0,DVDT_0
	movff	DSKIO_SEC_1,DVDT_1
	movff	DSKIO_SEC_2,DVDT_2
	movff	DSKIO_SEC_3,DVDT_3		; 32 bits
	movff	PRODL,DVSR_0
	movff	PRODH,DVSR_1
	clrf	DVSR_2
	clrf	DVSR_3
	call	DIV32
	movf	DVDT_0,W,BANKED
	movwf	CYLPRM,ACCESS			; cylinder LOW
	movf	DVDT_1,W,BANKED
	movwf	SPDCOU,ACCESS			; cylinder HIGH

	; now calculate head Hd=(LS%(Sec*Hd))/Sec
	movlw	0x10
	andwf	DRVPRM,1,ACCESS			; reset head parameter in the DRVPRM

	movff	RMDR_0,DVDT_0
	movff	RMDR_1,DVDT_1			; assume remainder was 16-bit
	movff	DSKIO_SpT,DVSR_0
	clrf	DVSR_1,BANKED
	call	DIV16
	movf	DVDT_0,W,BANKED
	andlw	0x0f					; max 15 heads
	iorwf	DRVPRM,1,ACCESS

	; process sector number
	incf	RMDR_0,W,BANKED			; remainder +1 (FDC sector number starts with 1)
	movwf	SECPRM,ACCESS
	movlb	0x00					; return back to bank 0

	; now ready to write HDD registers
	movf	DRVPRM,W,ACCESS
	call	WriteHDDDrvHead
	call	WaitHDDReady

	movf	CYLPRM,W,ACCESS
	call	WriteHDDCylLow
	movf	SPDCOU,W,ACCESS
	call	WriteHDDCylHigh
	movf	SECPRM,W,ACCESS
	call	WriteHDDSectorNumber
	movlw	0x01
	call	WriteHDDSectorCount

	; waiting for respective HDD to become ready
	call	WaitHDDReady

	movf	HOSTCMD,W,ACCESS
	decf	WREG,W,ACCESS
	bz		DIO_HDD_ReadOp
	movlw	0x30					; write with retries
	bra		DIO_Hdd_CmdSetup
DIO_HDD_ReadOp:
	movlw	0x20
DIO_Hdd_CmdSetup:
	call	WriteHDDCommand			; issue command to the HDD

	; wait for data ready
	clrf	TIMOUT3,ACCESS
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS			; long timeout = 0x1000000 loops

HDDReadWaitDRQ:
	call	ReadHDDStatus

	btfsc	LATA,0,ACCESS			; ERR bit - error occured
	bra		DIO_HDD_ErrDecode

	movf	LATA,W,ACCESS
	andlw	0x88
	sublw	0x08					; BUSY (bit 7)=0, DRQ (bit 3)=1
	bz		HDD_SendReceiveData

	decfsz	TIMOUT1,1,ACCESS
	bra		HDDReadWaitDRQ
	decfsz	TIMOUT2,1,ACCESS
	bra		HDDReadWaitDRQ
	decfsz	TIMOUT3,1,ACCESS
	bra		HDDReadWaitDRQ
	bra		DIO_HDDTimeout

HDD_SendReceiveData:
	clrf	FMTPR1,ACCESS			; counter clear, 256 counts

	decfsz	HOSTCMD,W,ACCESS
	bra		WriteHDDDataData

ReadHDDDataData:
	call	ReadHDDDataReg
	movf	LATA,W,ACCESS
	movwf	POSTINC0,ACCESS
	movf	LATB,W,ACCESS
	movwf	POSTINC0,ACCESS
	decfsz	FMTPR1,1,ACCESS
	bra		ReadHDDDataData
	bra		HDD_EndOp

WriteHDDDataData:
	movf	POSTINC0,W,ACCESS
	movwf	LATA,ACCESS
	movf	POSTINC0,W,ACCESS
	movwf	LATB,ACCESS
	call	WriteHDDDataReg
	decfsz	FMTPR1,1,ACCESS
	bra		WriteHDDDataData

HDD_EndOp:
	; waiting for respective HDD to become ready
	call	WaitHDDReady
	btfsc	LATA,5,ACCESS			; bit 5 of status register, DWF (Data Write Fault)
	bra		DIO_HDD_WriteFault

	btfsc	LATA,0,ACCESS			; bit 0 of status register, ERR (Error occured)
	bra		DIO_HDD_ErrDecode

	btfsc	LATA,3,ACCESS
	bra		HDD_EndOp				; DRQ is still active! go wait until it clears

	; *** check if there're more sectors to read/write
	incf	DIO_NumSecProcessed,1,ACCESS
	movf	DIO_NumSecProcessed,W,ACCESS
	subwf	DSKIO_NumOfSec,W,ACCESS
	bnz		HDD_NextSector

	movlw	0xcc
	bra		HDD_ReturnStatus		; exit with success

HDD_NextSector:
	; *** no, not all sectors were processed
	incf	DSKIO_SEC_0,1,ACCESS
	clrf	WREG,ACCESS				; does not affect C flag
	addwfc	DSKIO_SEC_1,1,ACCESS
	addwfc	DSKIO_SEC_2,1,ACCESS
	addwfc	DSKIO_SEC_3,1,ACCESS	; next logical sector (32 bits)
	bra		DIO_HDD_NextFullLoop	; go back with logical sector +1, sector count-1,
									; and FSR0 shifted by 0200h
DIO_HDD_WriteFault:
	movlw	0x04					; "CRC error"
	bra		HDD_ReturnStatus
DIO_HDD_ErrDecode:
	; *** error occured - need to read error register
	; and return proper status code to the host
	call	ReadHDDError
;	movwf	TMP,ACCESS
	movlw	0x90					; BBK, INDF, TK0NF, AMNF
	andwf	LATA,W,ACCESS
	bnz		DIO_HDD_NF				; go "record not found"
	movlw	0x40					; UNC
	andwf	LATA,W,ACCESS
	bnz		DIO_HDD_WriteFault		; go "CRC error"
DIO_HDDTimeout:
	movlw	0x02					; "not ready"
	bra		HDD_ReturnStatus

DIO_HDD_NF:
	movlw	0x08					; record not found
	bra		HDD_ReturnStatus

; --- Waits for HDD become ready (BSY=0, RDY=1)
WaitHDDReady:
	clrf	TIMOUT3,ACCESS
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS			; long timeout = 0x1000000 loops

HDDReadWaitReady:
	call	ReadHDDStatus
	movf	LATA,W,ACCESS
	andlw	0xC0
	sublw	0x40					; BUSY (bit 7)=0, RDY (bit 6)=1
	bz		HDDRetReadWaitReady

	decfsz	TIMOUT1,1,ACCESS
	bra		HDDReadWaitReady
	decfsz	TIMOUT2,1,ACCESS
	bra		HDDReadWaitReady
	decfsz	TIMOUT3,1,ACCESS
	bra		HDDReadWaitReady
	bra		DIO_HDDTimeout
HDDRetReadWaitReady:
	return

;*************************** Seek and recalibrate commands **********************************
;--------------------- Seek to specified cylinder and remember it ---------------------------
; in: (CURDRV) = _physical_ drive number, (CYLPRM) = cylinder to seek to
; out: CRCYLx value to identify current head position of the _physical_ drive
; out: FDCERR
SeekCommand:
	call	ReadDiskChange
	btfsc	TEMPREQ,7,ACCESS
	call	DiskChangeFlagSet		; set disk change flag for specified drive for DSKCHG sub

	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x0F
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS			; access _physical_ drive
	call	AddToFDCQueue
	movf	CYLPRM,W,ACCESS
	call	SetPhyDriveCurCyl		; update position of the _physical_ drive's heads
	call	AddToFDCQueue
	call	RunCommandNoResult

Check_Seek_Err:
	movlw	0xff
	btfsc	FDCERR,0,ACCESS
	call	SetPhyDriveCurCyl		; set current cylinder of the physical drive to 0xff (unknown)
	return

;------------------------------------ Recalibrate drive --------------------------------------
; in: (CURDRV) = _physical_ drive number
; out: CRCYLx=0 for the _physical_ drive
; out: FDCERR
Recalibrate:
	clrf	WREG,ACCESS
	call	SetPhyDriveCurCyl		; update position of the _physical_ drive'e heads = 0

	; *** now we wii perform recalibration - using
	; TWO Recalibrate commands. It is required if heads are
	; above track #77

	; Recalibrate drive, attempt #1
	call	DoRecalibrate
	btfss	FDCERR,0,ACCESS
	bra		RecalibrateSuccessful

	; Recalibrate drive, attempt #2
	call	DoRecalibrate
	btfsc	FDCERR,0,ACCESS
	bra		Check_Seek_Err			; second attempt unsuccessful - error

RecalibrateSuccessful:
	; now we issue "SENSE DRIVE STATUS" command, and if track 0 signal (bit 4 of ST3)
	; is not set, drive is not present or defective
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue			; Sense drive status FDC command
	movf	CURDRV,W,ACCESS
	call	AddToFDCQueue
	call	RunCommand

	btfsc	FDCERR,0,ACCESS
	bra		Check_Seek_Err

	btfss	CMDRLOC,4,ACCESS		; bit 4 of ST3, "TRACK 0", is not set
	setf	FDCERR,ACCESS
	bra		Check_Seek_Err

; Issue Recalibrate command without any trailing error checking
DoRecalibrate:
	; Recalibrate drive, attempt #1
	call	NewFDCRequest
	movlw	0x07				; RECALIBRATE command
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS
	call	AddToFDCQueue
	call	RunCommandNoResult
	return

; --- Sets current head position (cylinder#) of the currently operated _physical_ drive ----
; in: WREG=new cylinder number
SetPhyDriveCurCyl:
	btfss	CURDRV,0,ACCESS
	goto	CurCylDrv0
	movwf	CURCYL1,ACCESS
	return
CurCylDrv0:
	movwf	CURCYL0,ACCESS
	return

; --- Gets current head position (cylinder#) of the currently operated _physical_ drive ----
; out: WREG=current cylinder number
GetPhyDriveCurCyl:
	btfss	CURDRV,0,ACCESS
	bra		GetCurCylDrv0
	movf	CURCYL1,W,ACCESS
	return
GetCurCylDrv0:
	movf	CURCYL0,W,ACCESS
	return
;************************** End of Seek and recalibrate commands *****************************




;--------------------------------------------------------------------------------------------

; in: CURDRV = current drive ID (bit 0)
; out: none
DiskChangeFlagSet:
	btfss	CURDRV,0,ACCESS
	bra		FlagSetDrv0

	; drive 1 disk change flag setting
	bsf		DSK1CHGFLG,0,ACCESS
	return
FlagSetDrv0:
	bsf		DSK0CHGFLG,0,ACCESS
	return

; in: CURDRV = current drive ID (bit 0)
; out: none
DiskChangeFlagReset:
	btfss	CURDRV,0,ACCESS
	bra		FlagResetDrv0

	; drive 1 disk change flag setting
	clrf	DSK1CHGFLG,ACCESS
	return
FlagResetDrv0:
	clrf	DSK0CHGFLG,ACCESS
	return

; in: CURDRV = current drive ID (bit 0)
; out: W = flag (bit 0)
DiskChangeFlagGet:
	btfss	CURDRV,0,ACCESS
	bra		FlagGetDrv0

	; drive 1 disk change flag setting
	movf	DSK1CHGFLG,W,ACCESS
	return
FlagGetDrv0:
	movf	DSK0CHGFLG,W,ACCESS
	return

DecodeError:
	; Errors returned:
	; [0 = write protected] - not returned here, needs sense status command
	; [2 = not ready] - not returned here
	; 4 = crc error
	; 6 = seek error
	; 8 = record not found
	; 10 = write fault
	; 12 = other error
	; 0xCC = no error

	btfsc	FDCERR,0,ACCESS
	bra		FDDNotReady				; most probably timeout occured, sense data returned by FDC may not be valid

	; ST0
	btfsc	CMDRLOC,4,ACCESS		; first byte of the result data, ST0, bit 4 = "CHECK EQUIPMENT"
	bra		SeekError

	; ST1
	btfsc	CMDRLOC+1,5,ACCESS		; second byte of the result data, ST1, bit 5 = "DATA ERROR"
	bra		CrcError
	btfsc	CMDRLOC+1,2,ACCESS		; second byte of the result data, ST1, bit 2 = "NO DATA"
	bra		RecordNotFound
	btfsc	CMDRLOC+1,4,ACCESS		; second byte of the result data, ST1, bit 4 = "OVERRUN"
	bra		FDDNotReady
	btfsc	CMDRLOC+1,1,ACCESS		; second byte of the result data, ST1, bit 1 = "NOT WRITABLE"
	bra		WriteFault
	btfsc	CMDRLOC+1,0,ACCESS		; second byte of the result data, ST1, bit 0 = "MISSING ADDRESS MARK"
	bra		RecordNotFound
;	btfsc	0x61,7					; second byte of the result data, ST1, bit 7 = "END OF CYLINDER"
;	goto	RecordNotFound

	; ST2
	btfsc	CMDRLOC+2,5,ACCESS		; second byte of the result data, ST1, bit 5 = "DATA ERROR"
	bra		CrcError
	btfsc	CMDRLOC+2,6,ACCESS		; second byte of the result data, ST1, bit 6 = "CONTROL MARK"
	bra		RecordNotFound
	btfsc	CMDRLOC+2,4,ACCESS		; second byte of the result data, ST1, bit 6 = "WRONG CYLINDER"
	bra		SeekError
	btfsc	CMDRLOC+2,1,ACCESS		; second byte of the result data, ST1, bit 6 = "BAD CYLINDER"
	bra		SeekError
	btfsc	CMDRLOC+2,4,ACCESS		; second byte of the result data, ST1, bit 6 = "NISSING ADDRESS MARK"
	bra		RecordNotFound

	retlw	0xCC					; no error
FDDNotReady:
	retlw	0x02
SeekError:
	retlw	0x06
RecordNotFound:
	retlw	0x08
CrcError:
	retlw	0x04
WriteFault:
	retlw	0x0A

DSKCHG:
	movlw	0xAB
	call	WaitForHostRead			; return 0xAB to the host - command has started
	btfss	STATUS,C,ACCESS			; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun			; in this case return 0xAA

	call	WaitForHostWrite		; and get first parameter: drive # (0=A:)
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	andlw	0x01					; drive number
;	movwf	CURDRV
	call	CheckSingleDrive		; populates CURHOSTDRV and CURDRV
;	movwf	DRVPRM,ACCESS
	call	MotorOnSelect
	call	ReadDiskChange			; check FDD's disk change bit
	call	MotorOff

	btfsc	TEMPREQ,7,ACCESS
	bra		DSKChanged

	call	DiskChangeFlagGet
	btfsc	WREG,0,ACCESS
	bra		DSKChanged

	movlw	0x01					; disk unchanged
	bra		RetChg
DSKChanged:
	call	DiskChangeFlagReset
	; now clear the disk change bit by seeking to cylinder 10 first, and the cylinder 0
;	call	SoftReset
	call	MotorOnSelect
	movlw	0x01
	movwf	CYLPRM,ACCESS
	call	SeekCommand				; Cylinder 1
	clrf	CYLPRM,ACCESS
	call	SeekCommand				; Cylinder 0
	call	MotorOff
	movlw	0xFF					; disk changed

RetChg:
	movwf	COUVAL,ACCESS
	movlw	0xAF
	call	WaitForHostRead				; return 0xAF to indicate that results are ready
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	COUVAL,W,ACCESS
	call	WaitForHostRead			; return value
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop

; ---------------------------------------
; reads disk change attribute from LDCR - currently selected drive
; returns: TEMPREQ=LDCR byte read
ReadDiskChange:
	; read LDCR, bit 7
	setf	PORTA,ACCESS
	clrf	PORTE,ACCESS			; activate /LDCR
	movlw	0xF8					; /LDOR=1, A=0, /CS=1: FDC control register selected
	movwf	PORTC,ACCESS
	bcf		LATC,3,ACCESS			; /RDI activated (important: do not use PORTC as bcf is rmw operation)
	setf	TRISA,ACCESS			; put port A into input mode
	nop								; /RDI impulse duration before data read: 187.5ns (>90ns Trd, as per datasheet)
	movf	PORTA,W,ACCESS			; wait 62.5*3=187.5 (>90 which is minimal for read signal to be active)
	clrf	TRISA,ACCESS			; set port A back to output mode
	bsf		LATC,3,ACCESS			; /RDI deactivated (important: do not use PORTC as bcf is rmw operation)
	setf	PORTC,ACCESS			; release /RDI
	setf	PORTE,ACCESS			; release /LDCR
	movwf	TEMPREQ,ACCESS

	; set disk change flag if diskchange bit is set
	btfsc	TEMPREQ,7,ACCESS
	call	DiskChangeFlagSet		; set disk change flag for DSKCHG sub
	return
; ---------------------------------------

DSKFMT:
	movlw	0xAB					; command accepted: now stage 2 to get first parameter from host
	call	WaitForHostRead			; return value
	btfss	STATUS,C,ACCESS			; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun			; in this case return 0xAA

	call	WaitForHostWrite
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	movwf	DRVPRM,ACCESS
	andlw	0x01
;	movwf	CURDRV					; current drive
	call	CheckSingleDrive

	call	WaitForHostWrite		; get second: format parameter
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command
	movwf	FMTPRM,ACCESS			; format parameter: bits 1/0: 00 = Std1.44,
									; 01=Comp1.44, 10=Lgcy720, 11=invalid
									; bit 2: bootsector DOS1 (0)/DOS2 (1)

	call	WaitForHostWrite		; get third: confirmation code
	btfss	STATUS,C,ACCESS			; A0 address line state during host request
	goto	NoCommandToRun			; access to port base+1 is not assumed this time, aborting MC command

	sublw	0xEC					; pass word
	btfss	STATUS,Z,ACCESS
	goto	NoCommandToRun

	movlw	0xAE
	call	WaitForHostRead			; return value to indicate that FORMAT command has started
	btfss	STATUS,C,ACCESS			; address A0 is set to 1 (reset MC operation process)
	goto	NoCommandToRun			; in this case return 0xAA

	; set drive speed
	call	Set144SpeedVals
	btfsc	FMTPRM,1,ACCESS			; bit 1 set: Legacy720 (10b)
	call	Set720SpeedVals
	call	InitDrive				; initialize drive (CURDRV) to format disk in, and sets speeds
	call	DiskChangeFlagSet		; Set disk change flag for the current drive in order
									; for the host to perform full re-read of disk structures
									; on the next madia access
	clrf	CYLPRM,ACCESS			; init cylinder#=0
	clrf	SECPRM,ACCESS			; init head#=0

	; checking for diskette write protection - issuing SENSE DRIVE STATUS command
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x04
	call	AddToFDCQueue
	movf	DRVPRM,W,ACCESS
	call	AddToFDCQueue
	call	RunCommand

	btfss	FDCERR,0,ACCESS
	goto	FmtWPReady

	movlw	0x02
	goto	DSKFMT_ReturnStatus		; "not ready"

FmtWPReady:
	btfss	CMDRLOC,6,ACCESS		; if bit 6 of ST3=1 then disk is write protected
	bra		FormatLoop

	; diskette is write protected
	clrf	WREG,ACCESS				; WRITE PROTECT error
	goto	DSKFMT_ReturnStatus

FormatLoop:
	; proceed with seek to the specified cylinder
	call	SeekCommand
	btfss	FDCERR,0,ACCESS
	bra		FormatSeekOk
Fmt_SeekError:
	movlw	0x06			; "seek error"
	goto	DSKFMT_ReturnStatus
FormatSeekOk:
	; *** set buffer pointer
	movlw	0x01
	movwf	FSR1H
	clrf	FSR1L					; buffer start is at 0x100
	; *** issue format command
	call	NewFDCRequest
	movlw	0x4D					; FORMAT TRACK command, MFM mode
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS
	iorwf	SECPRM,W,ACCESS			; OR with head number (in bit 2)
	call	AddToFDCQueue
	movlw	0x02					; Sector size = 512 bytes
	call	AddToFDCQueue
	movf	CURSPT,W,ACCESS			; Sectors per track (current value)
	call	AddToFDCQueue
	movf	CURGPL2,W,ACCESS		; gap length 2 for formatting
	call	AddToFDCQueue
	movlw	0xE5					; Filler byte for data
	call	AddToFDCQueue

	movf	CYLPRM,W,ACCESS
	call	FillBufForFormat		; build tables to format specified track, should be run
									; AFTER format command buffer is filled with data (uses byte at offset 0x51)
									; it fills data for 18 consequtive sectors, even if we format 9 SpT
	call	RunCommandWithData
	btfss	FDCERR,0,ACCESS
	bra		TrackFormatOk

FormatRetWriteFaultErr:
	movlw	0x0A					; "write fault"
	goto	DSKFMT_ReturnStatus

TrackFormatOk:
	; *** next head
	btg		SECPRM,2,ACCESS			; next head
	tstfsz	SECPRM,ACCESS
	bra		FormatSeekOk			; same cylinder, another head (1)

	; *** next cylinder
	incf	CYLPRM,1,ACCESS
	movf	CYLPRM,W,ACCESS
	sublw	0x50					; track 79 is the last one
	bnz		FormatLoop				; next cylinder

; --- Now writing control blocks: Boot sector, FAT and directory
	movf	FMTPRM,W,ACCESS
	andlw	0x03
	bnz		Compat144_720Format		; option not 00 => compatible format

	; standard 1.44 format (1 sectors per cluster, 9 FAT sectors, GR8BIT stack at 0C382h)
	movlw	0x21
	movwf	FMTPR1,ACCESS			; control block sector count (33)
	movlw	0x20					; first FAT is at sector 1
	movwf	FMTPR2,ACCESS			; first FAT counter
	movlw	0x17					; seconf FAT is at sector 10 (=1+9)
	movwf	FMTPR3,ACCESS			; second FAT counter
	bra		ProceedWrite

Compat144_720Format:
	; we are here for options 1, 2, and 3 (3=invaild) former is compatible 1.44 and
	; latter is compatible 720, which have same control structures' allocation
	; boot=0, fat1=1, fat2=4, dir=7, data=14
	; in both cases (SpT=18 and SpT=9) structure data located within cylinder 0

	; Compatible format (1 or 4 sectors per cluster, 3 FAT sectors, 720K compatible stack location)
	movlw	0x0E
	movwf	FMTPR1,ACCESS			; control block sector count (33)
	movlw	0x0D					; first FAT is at sector 1
	movwf	FMTPR2,ACCESS			; first FAT counter
	movlw	0x0A					; second FAT is at sector 4 (=1+3)
	movwf	FMTPR3,ACCESS			; second FAT counter
	bra		ProceedWrite

ProceedWrite:
	movf	FMTPR1,W,ACCESS			; sector count (0x21 for standard and 0x0E for compatible)
	movwf	FMTTMP1,ACCESS			; current logical sector number in countdown mode
	clrf	CYLPRM,ACCESS			; current cylinder#
	movlw	0x01
	movwf	SECPRM,ACCESS			; currect sector number
	movf	CURDRV,W,ACCESS
	movwf	DRVPRM,ACCESS			; head bit in CURDRV = 0 (head 0)

WriteBootFat:
	; proceed with seek to the specified cylinder
	call	SeekCommand
	btfsc	FDCERR,0,ACCESS
	bra		Fmt_SeekError

	call	ClearBufForFormat

	movf	FMTPR1,W,ACCESS			; sector count (0x21 for standard and 0x0E for compatible)
	subwf	FMTTMP1,W,ACCESS
	bnz		NotBootSector

	; *** boot sector
	movlw	0x01
	movwf	FSR0H,ACCESS
	clrf	FSR0L,ACCESS			; set up indirect pointer to the buffer - 0x100

	movlw	0xb7
	movwf	FMTTMP2,ACCESS			; byte count in boot sector data

; prepare flash memory table access
	btfsc	FMTPRM,2,ACCESS			; test for DOS1/DOS2 boot sector option: 0=DOS1, 1=DOS2
	bra		BootDOS2

	movlw	UPPER BootSectDOS1
	movwf	TBLPTRU,ACCESS
	movlw	HIGH BootSectDOS1
	movwf	TBLPTRH,ACCESS
	movlw	LOW BootSectDOS1
	movwf	TBLPTRL,ACCESS
	bra		BootFill

BootDOS2:
	movlw	UPPER BootSectDOS2
	movwf	TBLPTRU,ACCESS
	movlw	HIGH BootSectDOS2
	movwf	TBLPTRH,ACCESS
	movlw	LOW BootSectDOS2
	movwf	TBLPTRL,ACCESS

BootFill:
	movf	FMTTMP2,W,ACCESS
	sublw	0xb8
	movwf	FMTTMP3,ACCESS

	tblrd*+
	movf	TABLAT,W,ACCESS			; read boot sector byte from flash ROM
	movwf	POSTINC0,ACCESS			; save boot sector byte to the buffer

	decfsz	FMTTMP2,1,ACCESS
	bra		BootFill

	; now fill boot sector with proper disk configuration data
	; combinations of bits 0 and 1 in FMTPRM:
	; 00 = standard 1.44
	; 01 = compatible 1.44
	; 10 = legacy 720
	; 11 = illegal
	movf	FMTPRM,W,ACCESS
	andlw	0x03
	bnz		FmtOption01_10_11

FmtOption00:
	movlb	1
	movlw	0x01			; Standard config: 1 sector per cluster, 224 dir entries, 9 sectors per FAT
	movwf	0x0D,BANKED		; cluster size
	movlw	0xE0
	movwf	0x11,BANKED		; dir entries
	movlw	0x09
	movwf	0x16,BANKED		; size of FAT
	movlb	0
	bra		NotFAT

FmtOption01_10_11:
	decfsz	WREG,1,ACCESS	; illegal combination: go for standard disk header
	bra		Compat720

	movlb	1
	movlw	0x04			; Compatible config 1.44M: 4 sectors per cluster, 112 dir entries, 3 sectors per FAT
	movwf	0x0D,BANKED		; cluster size
	movlw	0x70
	movwf	0x11,BANKED		; dir entries
	movlw	0x03
	movwf	0x16,BANKED		; size of FAT
	movlb	0
	bra		NotFAT

Compat720:
	movlb	1
	movlw	0x02			; Compatible config 720K: 2 sectors per cluster, 112 dir entries, 3 sectors per FAT
	movwf	0x0D,BANKED		; cluster size
	movlw	0x70
	movwf	0x11,BANKED		; dir entries
	movlw	0x03
	movwf	0x16,BANKED		; size of FAT
	movlw	0xA0
	movwf	0x13,BANKED
	movlw	0x05
	movwf	0x14,BANKED		; sectors on disk (0x5A0)
	movlw	0xF9
	movwf	0x15,BANKED		; 720K media descriptor (0xF9)
	movlw	0x09
	movwf	0x18,BANKED		; number of sectors per track (0x9)
	movlb	0
	bra		NotFAT

NotBootSector:
	movf	FMTPR2,W,ACCESS			; first FAT pointer count (0x20 for standard and 0x0D for compatible)
	subwf	FMTTMP1,W,ACCESS
	bz		FATSector
	movf	FMTPR3,W,ACCESS			; second FAT pointer count (0x17 for standard and 0x0A for compatible)
	subwf	FMTTMP1,W,ACCESS
	bnz		NotFAT

FATSector:
	; select propert media descriptor for FAT copies
	movlw	0xF0
	btfsc	FMTPRM,1,ACCESS
	movlw	0xF9					; 10/11 = Legacy720, descriptor 0F9h

	movlb	1
	movwf	0x00,BANKED
	setf	0x01,BANKED
	setf	0x02,BANKED			; FAT "F0 FF FF ..."
	movlb	0

NotFAT:
	; *** set buffer pointer
	movlw	0x01
	movwf	FSR1H
	clrf	FSR1L					; buffer start is at 0x100

	; *** write sector to the media
	call	NewFDCRequest			; clear MC FDC queue
	movlw	0x45
	call	AddToFDCQueue			; FDC write command
	movf	DRVPRM,W,ACCESS
	call	AddToFDCQueue
	movf	CYLPRM,W,ACCESS
	call	AddToFDCQueue
	rrncf	DRVPRM,W,ACCESS
	rrncf	WREG,1,ACCESS
	andlw	0x01
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS
	call	AddToFDCQueue
	movlw	0x02
	call	AddToFDCQueue
	movf	SECPRM,W,ACCESS			; last sector# on the track to transfer (the same as sector# to read)
	call	AddToFDCQueue
	movf	CURGPL1,W,ACCESS		; GAP1 default size
	call	AddToFDCQueue
	movlw	0xFF					; Data length (default 0xFF)
	call	AddToFDCQueue
	call	RunCommandWithData

	btfsc	FDCERR,0,ACCESS
	goto	FormatRetWriteFaultErr	; "WRITE FAULT"

	; *** next sector
	incf	SECPRM,1,ACCESS

	; setting max sector number +1
	movlw	0x13					; 18 sectors + 1
	btfsc	FMTPRM,1,ACCESS
	movlw	0x0A					; 9 sectors + 1

	subwf	SECPRM,W,ACCESS
	bnz		NoNewHead

	movlw	0x01
	movwf	SECPRM,ACCESS			; sector 1 of the new head

	; *** next head here
	btg		DRVPRM,2,ACCESS
	btfsc	DRVPRM,2,ACCESS
	bra		NoNewHead

	; *** next track here
	incf	CYLPRM,1,ACCESS

NoNewHead:
	decfsz	FMTTMP1,1,ACCESS
	bra		WriteBootFat

	; format ended successfully
	movlw	0xCC

DSKFMT_ReturnStatus:
	; *** return control data and exit
	movwf	SECPRM,ACCESS			; preserve return byte
	call	MotorOff

	movlw	0xAF
	call	WaitForHostRead			; returning 0xAF - result phase
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command

	movf	SECPRM,W,ACCESS				; restore return byte
	call	WaitForHostRead
	btfss	STATUS,C,ACCESS				; A0 address line state during host request
	goto	NoCommandToRun				; access to port base+1 is not assumed this time, aborting MC command
	goto	HostWaitMainLoop

; DOS1 boot sector (no "VOL_ID" tag) - disk always boots into DOS1 mode
BootSectDOS1:
	db		0xeb,0xfe,0x90,0x47,0x52,0x38,0x42,0x49,0x54,0x31,0x30,0x00,0x02,0x00,0x01,0x00		; 0x00-0x0F
		;	  EB   FE   90  'G'  'R'  '8'  'B'  'I'  'T'  '1'  '0' BytesPSec  SpC  ResvSect
	db		0x02,0x00,0x00,0x40,0x0b,0xf0,0x00,0x00,0x12,0x00,0x02,0x00,0x00,0x00,0xd0,0xed
		;	nFAT RootDirEn TotalSect  MID SecPerFAT SecPerTra  NumHeads  HidSect
	db		0x53,0x59,0xc0,0x32,0xd0,0xc0,0x36,0x56,0x23,0x36,0xc0,0x31,0x1f,0xf5,0x11,0xab
	db		0xc0,0x0e,0x0f,0xcd,0x7d,0xf3,0x3c,0xca,0x63,0xc0,0x11,0x00,0x01,0x0e,0x1a,0xcd
	db		0x7d,0xf3,0x21,0x01,0x00,0x22,0xb9,0xc0,0x21,0x00,0x3f,0x11,0xab,0xc0,0x0e,0x27
	db		0xcd,0x7d,0xf3,0xc3,0x00,0x01,0x58,0xc0,0xcd,0x00,0x00,0x79,0xe6,0xfe,0xfe,0x02
	db		0xc2,0x6a,0xc0,0x3a,0xd0,0xc0,0xa7,0xca,0x22,0x40,0x11,0x85,0xc0,0xcd,0x77,0xc0
	db		0x0e,0x07,0xcd,0x7d,0xf3,0x18,0xb4,0x1a,0xb7,0xc8,0xd5,0x5f,0x0e,0x06,0xcd,0x7d
	db		0xf3,0xd1,0x13,0x18,0xf2,0x42,0x6f,0x6f,0x74,0x20,0x65,0x72,0x72,0x6f,0x72,0x0d
	db		0x0a,0x50,0x72,0x65,0x73,0x73,0x20,0x61,0x6e,0x79,0x20,0x6b,0x65,0x79,0x20,0x66
	db		0x6f,0x72,0x20,0x72,0x65,0x74,0x72,0x79,0x0d,0x0a,0x00,0x00,0x4d,0x53,0x58,0x44
	db		0x4f,0x53,0x20,0x20,0x53,0x59,0x53

; DOS2 boot sector (with "VOL_ID" tag) - can boot in DOS2 mode
BootSectDOS2:
	db		0xeb,0xfe,0x90,0x47,0x52,0x38,0x42,0x49,0x54,0x32,0x30,0x00,0x02,0x00,0x01,0x00		; 0x00-0x0F
		;	  EB   FE   90  'G'  'R'  '8'  'B'  'I'  'T'  '2'  '0' BytesPSec  SpC  ResvSect
	db		0x02,0x00,0x00,0x40,0x0b,0xf0,0x00,0x00,0x12,0x00,0x02,0x00,0x00,0x00,0x18,0x10
		;	nFAT RootDirEn TotalSect  MID SecPerFAT SecPerTra  NumHeads  HidSect
	db		0x56,0x4f,0x4c,0x5f,0x49,0x44,0x00,0x39,0x7d,0x03,0x63,0x00,0x00,0x00,0x00,0x00
		;	 'V'  'O'  'L'  '_'  'I'  'D'
	db		0xd0,0xed,0x53,0x6a,0xc0,0x32,0x72,0xc0,0x36,0x67,0x23,0x36,0xc0,0x31,0x1f,0xf5
	db		0x11,0xab,0xc0,0x0e,0x0f,0xcd,0x7d,0xf3,0x3c,0x28,0x26,0x11,0x00,0x01,0x0e,0x1a
	db		0xcd,0x7d,0xf3,0x21,0x01,0x00,0x22,0xb9,0xc0,0x21,0x00,0x3f,0x11,0xab,0xc0,0x0e
	db		0x27,0xcd,0x7d,0xf3,0xc3,0x00,0x01,0x69,0xc0,0xcd,0x00,0x00,0x79,0xe6,0xfe,0xd6
	db		0x02,0xf6,0x00,0xca,0x22,0x40,0x11,0x85,0xc0,0x0e,0x09,0xcd,0x7d,0xf3,0x0e,0x07
	db		0xcd,0x7d,0xf3,0x18,0xb8,0x42,0x6f,0x6f,0x74,0x20,0x65,0x72,0x72,0x6f,0x72,0x0d
	db		0x0a,0x50,0x72,0x65,0x73,0x73,0x20,0x61,0x6e,0x79,0x20,0x6b,0x65,0x79,0x20,0x66
	db		0x6f,0x72,0x20,0x72,0x65,0x74,0x72,0x79,0x0d,0x0a,0x24,0x00,0x4d,0x53,0x58,0x44
	db		0x4f,0x53,0x20,0x20,0x53,0x59,0x53


#include<dos_files_v2_0a.inc>


SetSpeeds:
	; set channel speed
	movf	CURDSPD,W,ACCESS
	movwf	PORTA,ACCESS
	clrf	PORTE,ACCESS		; activate /LDCR

	movlw	0xF8				; /LDOR=1, A=0, /CS=1: FDD controller operations register selected
	movwf	PORTC,ACCESS
	bcf		LATC,4,ACCESS		; /WRI activated (important: do not use PORTC as bcf is rmw operation)
	nop							; /WRI impulse width: 125ns (>60ns as per datasheet)
	bsf		LATC,4,ACCESS		; /WRI deactivated (important: do not use PORTC as bcf is rmw operation)
	setf	PORTC,ACCESS		; release /WRI
	setf	PORTA,ACCESS		; release data bus
	setf	PORTE,ACCESS		; release /LDCR

; --- Specify (global attribute about disk drives' modes and times)
	call	NewFDCRequest
	movlw	0x03				; SPECIFY command
	call	AddToFDCQueue
	movf	CURDSP1,W,ACCESS
	call	AddToFDCQueue
	movf	CURDSP2,W,ACCESS
	bcf		WREG,0,ACCESS		; set DMA bit on
	call	AddToFDCQueue
	call	RunCommand
	return

; --- SOFT RESET OF FDD CONTROLLER ---
SoftReset:
	; Soft resetting FDD controller: LDOR=0/1, WR=1/0/1, 8 bit data bus
	clrf	PORTA,ACCESS		; SRST=0 (reset), DMAEN=0 (no DMA/IRQ)
	movlw	0x78				; /LDOR=0, A=0, /CS=1: FDD controller operations register selected
	movwf	PORTC,ACCESS
	bcf		LATC,4,ACCESS		; /WRI activated (important: do not use PORTC as bcf is rmw operation)
	nop							; /WRI impulse width: 125ns (>60ns as per datasheet)
	bsf		LATC,4,ACCESS		; /WRI deactivated (important: do not use PORTC as bcf is rmw operation)
	setf	PORTC,ACCESS
	setf	PORTA,ACCESS		; setting data bus to high state

	movlw	0xFF
	call	Wait_ms				; wait 255 ms

	call	ReadMainStatRegFDC

	movlw	0x04				; release reset, DMAEN=0 (disabled)
	movwf	PORTA,ACCESS
	movwf	DORVAL,ACCESS
	movlw	0x78				; /LDOR=0, A=0, /CS=1: FDD controller operations register selected
	movwf	PORTC,ACCESS
	bcf		LATC,4,ACCESS		; /WRI activated (important: do not use PORTC as bcf is rmw operation)
	nop							; /WRI impulse width: 125ns (>60ns as per datasheet)
	bsf		LATC,4,ACCESS		; /WRI deactivated (important: do not use PORTC as bcf is rmw operation)
	setf	PORTC,ACCESS
	setf	PORTA,ACCESS
; --- RESET COMPLETE ---

	movlw	0xFF
	call	Wait_ms				; wait 255 ms

	call	ReadMainStatRegFDC
	call	SenseInterruptCmd

	setf	TIMOUT1,ACCESS		; timeout: 255 loops
CheckAgain:
	call	ReadMainStatRegFDC	; read main status register
	btfsc	MAINSTAT,4,ACCESS
	bra		NotReady			; go there if bit 4 = 1 (command is executing)
	btfss	MAINSTAT,7,ACCESS
	bra		NotReady			; go there if bit 7 = 0 (FDC is not ready for data exchange)
	btfss	MAINSTAT,6,ACCESS
	bra		ReadyToWrite		; go there if bit 6 = 0 (ok to write to FDC)
	; FDC has something to say here (do not put port A to input as it does not matter what FDC is saying)
	setf	PORTA,ACCESS		; port A - high level
	movlw	0xD9				; /FDDCS(0)+A0(1) = data register selected
	movwf	PORTC,ACCESS
	bcf		LATC,3,ACCESS		; /RDI activated (important: do not use PORTC as bcf is rmw operation)
	nop							; /RDI impulse width: 125ns (>90ns as per datasheet)
	bsf		LATC,3,ACCESS		; /RDI deactivated (important: do not use PORTC as bcf is rmw operation)
	setf	PORTC,ACCESS
NotReady:
	decfsz	TIMOUT1,1,ACCESS
	bra		CheckAgain
	bra		SoftReset		; reset again after timeout

ReadyToWrite:
; --- Set up data rate - writing to CR

	call	SetSpeeds
	return

; ----------------------- DMA enable and disable routines ----------------------
EnableDMA:
	; *** setting PORTB mode and output value
	movlw	0x04
	movwf	LATB,ACCESS			; /DACK and TC lines are inactive
	movlw	0x03
	movwf	TRISB,ACCESS		; DI8 and DI9 are inputs, other pins are outputs

	bsf		DORVAL,3,ACCESS		; Enable DMA
	bra		DORWrite
DisableDMA:
	bcf		DORVAL,3,ACCESS		; Disable DMA
	call	DORWrite

	setf	LATB,ACCESS			; outputs 0xff
	clrf	TRISB,ACCESS		; port B is output
	return

DORWrite:
	movf	DORVAL,W,ACCESS
	movwf	PORTA,ACCESS
	movlw	0x78				; /LDOR=0, A=0, /CS=1: FDD controller operations register (AT/EISA mode) selected
	movwf	PORTC,ACCESS
	bcf		LATC,4,ACCESS		; /WRI activated (important: do not use PORTC as bcf is rmw operation)
	nop							; /WRI impulse width: 125ns (>60ns as per datasheet)
	bsf		LATC,4,ACCESS		; /WRI deactivated (important: do not use PORTC as bcf is rmw operation)
	setf	PORTC,ACCESS
	setf	PORTA,ACCESS
	return

; ------------------------ Initialize drive ------------------------
; in:	CURDRV = current drive
; out:	RETERR = set if error occured at any of the init steps
InitDrive:
	; Turn motor on
	call	MotorOnSelect
	call	SetSpeeds

	; recalibrating drive
	call	Recalibrate
	btfsc	FDCERR,0,ACCESS
	bra		RetError

	; Seek to track 79
	call	NewFDCRequest
	movlw	0x0F				; SEEK command
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS
	call	AddToFDCQueue
	movlw	0x4F				; new cylinder number: 79
	call	AddToFDCQueue
	call	RunCommandNoResult	; no result phase
	btfsc	FDCERR,0,ACCESS
	bra		RetError

; --- Seek to track 0, drive 0
	call	NewFDCRequest
	movlw	0x0F				; SEEK command
	call	AddToFDCQueue
	movf	CURDRV,W,ACCESS
	call	AddToFDCQueue
	movlw	0x00				; new cylinder number: 00
	call	AddToFDCQueue
	call	RunCommandNoResult	; no result phase
	btfsc	FDCERR,0,ACCESS
	bra		RetError
	bcf		STATUS,C,ACCESS
	return

RetError:
	call	MotorOff	; start motor off countdown for current drive
	bsf		STATUS,C,ACCESS
	return
; ************************ Initialize drive ************************


; --------------- Run command --------------------
; in:	none
; out:	status bytes in bank 0 starting offset 0x40,
;		FDCRUNPOS=number of results read, FDCERR<>0 if error
RunCommandWithData:
	setf	CMDRUN,ACCESS		; set "command is running" flag so that host receive status "busy"
	call	EnableDMA
	call	MotorOnSelect
	clrf	FDCERR,ACCESS
	call	CommandPhase
	call	ExecutePhase
	call	ResultPhase
	call	MotorOff
	call	DisableDMA
	clrf	CMDRUN,ACCESS
	return
; *************** Run command ********************

; --------------- Execute phase --------------------
ExecutePhase:
	; *** pre-clearing temporary storage (for TC bit)
	clrf	TMP,ACCESS

	; *** clear byte counter
	clrf	RECLENL,ACCESS
	movlw	0x02
	movwf	RECLENH,ACCESS			; byte count in the record (0x200)

	; *** clear timeout counter
	clrf	TIMOUT3,ACCESS
	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS		; long timeout = 0x10000 loops

	; wait for execution phase start
ExecLoop:
	; *** reading DMA channel (DI9)
	btfsc	PORTB,1,ACCESS		; checking DMA line
	bra		DMADataAvailable

	; *** check for interrupt IRQ=1 (exec phase complete)
	btfsc	PORTB,0,ACCESS
	bra		ExecReturn			; execution phase finished: return

	; and wait in loop
	decfsz	TIMOUT1,1,ACCESS
	bra		ExecLoop
	decfsz	TIMOUT2,1,ACCESS
	bra		ExecLoop
	decfsz	TIMOUT3,1,ACCESS
	bra		ExecLoop

ExecError:
	; timeout occured
	call	SoftReset			; perform soft reset of the controller
ExecErrReturn:
	setf	FDCERR,ACCESS
ExecReturn:
	return

DMADataAvailable:
	; *** respond with /DACK and TC if it is set
	movff	TMP,LATB			; TMP has bit 2 reset (/DACK active)
								; and TC set appropriately
	btfsc	CMDWLOC,0,ACCESS		; which direction DMA works?
	bra		DMAWriteDir				; command's bit 0=1 => data will be supplied to FDC
	; DMA read from FDC
	call	ReadDataRegFDC		; read data register
	movwf	POSTINC1,ACCESS			; write data to buffer and increase FSR0
	bra		EndDMACycle
DMAWriteDir:
	; DMA write to FDC
	movf	POSTINC1,W,ACCESS
	call	WriteDataRegFDC

EndDMACycle:
	movlw	0x04
	movwf	PORTB,ACCESS		; deactivate /DMA and TC lines

	; *** clearing temporary storage (for TC bit)
	clrf	TMP,ACCESS

	; *** checking for last byte (to set TC active)
	movf	RECLENL,W,ACCESS
	sublw	0x02
	bnz		ExecLoop	;NotLastByte
	movf	RECLENH,W,ACCESS
	sublw	0x01
	bnz		ExecLoop	;NotLastByte

	; *** we are here when RECLENH:RECLENL=1:1
	; meaning last byte to transfer
	bsf		TMP,3,ACCESS		; raise bit 3 to set TC in next (last) loop
	bra		ExecLoop
; *************** Execute phase ********************

; --------------- Run command --------------------
; in:	none
; out:	status bytes in bank 0 starting offset 0x40,
;		FDCRUNPOS=number of results read, FDCERR<>0 if error
RunCommand:
	clrf	FDCERR,ACCESS
	setf	CMDRUN,ACCESS			; set "command is running" flag so that host receive status "busy"
	call	MotorOnSelect
	call	CommandPhase
	call	ResultPhase
	call	MotorOff
	clrf	CMDRUN,ACCESS
	return
; *************** Run command ********************

RunCommandNoResult:
	; send command (which has no result phase - seek and recalibrate)
	clrf	FDCERR,ACCESS
	setf	CMDRUN,ACCESS			; set "command is running" flag so that host receive status "busy"
	call	EnableDMA
	call	MotorOnSelect
	call	CommandPhase

	; now waiting for interrupt signal (very long timeout)
	clrf	TIMOUT3,ACCESS			; long timout!
	clrf	TIMOUT2,ACCESS			; long timout!
	clrf	TIMOUT1,ACCESS
IRQLoop:
	btfsc	PORTB,0,ACCESS			; bit 0 = IRQ line
	bra		InterruptOk
	decfsz	TIMOUT1,1,ACCESS
	bra		IRQLoop
	decfsz	TIMOUT2,1,ACCESS
	bra		IRQLoop
	decfsz	TIMOUT3,1,ACCESS
	bra		IRQLoop

	; timeout
	setf	FDCERR,ACCESS

InterruptOk:
	call	DisableDMA

	; checking source of interrupt, and condition
	call	SenseInterruptCmd

	movlw	0xe0
	andwf	CMDRLOC,W,ACCESS
	sublw	0x20
	btfss	STATUS,Z,ACCESS
	setf	FDCERR,ACCESS			; should be 001xxxxx for successful completion of command

	call	MotorOff
	call	DisableDMA
	clrf	CMDRUN,ACCESS
	return

; --------------- Reads data byte from the FDD controller --------------------
; in:	none
; out	data byte in W
ResultPhase:
	clrf	FSR0H,ACCESS			; indirect addressing bank 0
	movlw	CMDRLOC
	movwf	FSR0L,ACCESS			; low byte of the pointer
	clrf	FDCRUNPOS,ACCESS

	clrf	TIMOUT2,ACCESS			; long timout! ; 0x08
	clrf	TIMOUT1,ACCESS			; timeout = 0x800 loops
ReadLoop:
	call	ReadMainStatRegFDC
	btfsc	MAINSTAT,7,ACCESS		; 0x80, RQM
	bra		DataRegReady

	decfsz	TIMOUT1,1,ACCESS
	bra		ReadLoop
	decfsz	TIMOUT2,1,ACCESS
	bra		ReadLoop

	; timeout - exit
ReadErrorExit:
	setf	FDCERR,ACCESS
	return

DataRegReady:
	btfss	MAINSTAT,4,ACCESS		; 0x10, CB
	return							; result phase complete - return without error

	btfss	MAINSTAT,6,ACCESS		; DIO - should be 1, i.e. data ready for reading
	bra		ReadLoop

	call	ReadDataRegFDC
	movwf	POSTINC0,ACCESS			; write byte to the indirectly addressed location

	incf	FDCRUNPOS,1,ACCESS		; increment current pointer to the queue, increase pointer
	movf	FDCRUNPOS,W,ACCESS
	sublw	0x10
	bnz		ReadLoop				; keep filling buffer if no buffer overflow (16 bytes)
	goto	ReadErrorExit			; result buffer overrun (max 16 result bytes)
; *************** Reads data byte from the FDD controller ********************

; ----------------------- COMMAND PHASE ------------------------
; Writes queue contents to the FDC
CommandPhase:
	clrf	FSR0H,ACCESS			; indirect addressing bank 0
	movlw	CMDWLOC
	movwf	FSR0L,ACCESS			; low byte of pointer to the buffer
	clrf	FDCRUNPOS,ACCESS

	clrf	TIMOUT2,ACCESS
	clrf	TIMOUT1,ACCESS			; timeout = 0x800 loops

WriteLoop:
	call	ReadMainStatRegFDC
	btfsc	MAINSTAT,7,ACCESS		; 0x80, RQM
	bra		WriteRQMReady
CPhWait:
	decfsz	TIMOUT1,1,ACCESS
	bra		WriteLoop
	decfsz	TIMOUT2,1,ACCESS
	bra		WriteLoop

	; timeout - exit
WriteErrorExit:
	setf	FDCERR,ACCESS
	return

WriteRQMReady:
	btfsc	MAINSTAT,6,ACCESS
	bra		CPhWait					;WriteErrorExit		; DIO=1 means controller has unexpectedly something to say

	movf	POSTINC0,W,ACCESS		; read byte from the indirectly addressed location and increase pointer
	call	WriteDataRegFDC

	incf	FDCRUNPOS,1,ACCESS		; increment current pointer to the queue
	movf	FDCRUNPOS,W,ACCESS
	subwf	FDCQPOS,W,ACCESS
	bnz		WriteLoop
	return						; FDCERR is preserved
; *********************** COMMAND PHASE ************************

; ------------------- Reads FDC data register -----------------------
; returns data byte in W
ReadDataRegFDC:
	setf	PORTA,ACCESS		; port A - high level
	movlw	0xD9				; /FDDCS+A0(1) = data register selected
	movwf	PORTC,ACCESS
	bcf		LATC,3,ACCESS		; /RDI activated (important: do not use PORTC as bcf is rmw operation)
	setf	TRISA,ACCESS
	nop							; /RDI impulse duration before data read: 187.5ns (>90ns Trd, as per datasheet)
	movf	PORTA,W,ACCESS
	bsf		LATC,3,ACCESS		; /RDI deactivated (important: do not use PORTC as bcf is rmw operation)
	clrf	TRISA,ACCESS
	setf	PORTC,ACCESS
	movwf	DATAVAL,ACCESS
	return
; ******************* Reads FDC data register ***********************

; ------------------- Writes to FDC data register -----------------------
; in:	W = data byte to write
WriteDataRegFDC:
	movwf	PORTA,ACCESS
	movlw	0xD9				; /FDDCS+A0(1) = data register selected
	movwf	PORTC,ACCESS
	bcf		LATC,4,ACCESS		; /WRI activated (important: do not use PORTC as bcf is rmw operation)
	nop							; /WRI impulse width: 125ns (>60ns as per datasheet)
	bsf		LATC,4,ACCESS		; /WRI deactivated (important: do not use PORTC as bcf is rmw operation)
	setf	PORTC,ACCESS
	return
; ******************* Writes to FDC data register ***********************

; --------------- Reads FDC master status register ----------------------
; returns status byte in W
ReadMainStatRegFDC:
	setf	PORTA,ACCESS		; port A - high level
	movlw	0xD8				; /FDDCS+A0(0) = master register selected
	movwf	PORTC,ACCESS
	bcf		LATC,3,ACCESS		; /RDI activated (important: do not use PORTC as bcf is rmw operation)
	setf	TRISA,ACCESS
	nop							; /RDI impulse duration before data read: 187.5ns (>90ns Trd, as per datasheet)
	movf	PORTA,W,ACCESS
	clrf	TRISA,ACCESS
	bsf		LATC,3,ACCESS		; /RDI deactivated (important: do not use PORTC as bcf is rmw operation)
	setf	PORTC,ACCESS
	movwf	MAINSTAT,ACCESS
	return
; *************** Reads FDC master status register **********************

; -------------- Adds byte to the FDD command queue ---------------------
AddToFDCQueue:
	movwf	QTEMP,ACCESS
	clrf	FSR0H,ACCESS		; indirect addressing bank 0
	movf	FDCQPOS,W,ACCESS
	addlw	CMDWLOC
	movwf	FSR0L,ACCESS		; visible in all banks
	movf	QTEMP,W,ACCESS
	movwf	INDF0,ACCESS		; write to the indirectly addressed location
	incf	FDCQPOS,1,ACCESS
	return
; ************** Adds byte to the FDD command queue *********************

; ---------------- Clears FDD command/data queue ------------------------
NewFDCRequest:
	clrf	FDCQPOS,ACCESS
	return
; **************** Clears FDD command/data queue ************************

; ------------------------ SENSE INTERRUPT -------------------------------
SenseInterruptCmd:
	call	NewFDCRequest
	movlw	0x08				; SENSE INTERRUPT command
	call	AddToFDCQueue
	call	RunCommand
	return
; ************************ SENSE INTERRUPT *******************************

; ---------------- Wait millisecond, number of milliseconds in W -----------------
; W is preserved
Wait_ms:
	movwf	DELAY3,ACCESS
	clrf	DELAY1,ACCESS
Wait_ms2:
	movlw	0x14				; 0x0A on 32MHz
	movwf	DELAY2,ACCESS
Wait_ms1:
	decfsz	DELAY1,1,ACCESS
	bra		Wait_ms1
	decfsz	DELAY2,1,ACCESS
	bra		Wait_ms1
	decfsz	DELAY3,1,ACCESS
	bra		Wait_ms2
	return	
; **************** Wait millisecond, number of milliseconds in W *****************

; ---------------- Wait microsecond, number of microseconds in W -----------------
; 1+WREG*10 instruction cycles; for 64MHz one instruction takes 62.5ns, thus execution
; of 1 cycle is ~ equal to 625ns
Wait_us:
	movwf	DELAY1,ACCESS
Wait_us1:
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	nop
	decfsz	DELAY1,1,ACCESS
	bra		Wait_us1
	return
; **************** Wait microsecond, number of microseconds in W *****************

; ------------------ Turn motor on and clear turnoff counter ----------------
; in:	CURDRV = current drive #
MotorOnSelect:
	clrf	TMP,ACCESS
	movf	DORVAL,W,ACCESS
	btfss	CURDRV,0,ACCESS
	bra		Drv0

	btfsc	DORVAL,5,ACCESS		; bit of drive motor running
	incf	TMP,1,ACCESS		; TMP to change to 1	

	clrf	MOTBTO,ACCESS		; no drive B timeout - keep it running
	andlw	0xFC				; mask drive select bits
	iorlw	0x21				; turn drive B motor on and select drive B
	bra		WriteDOR

Drv0:
	btfsc	DORVAL,4,ACCESS		; bit of drive motor running
	incf	TMP,1,ACCESS		; TMP to change to 1	

	clrf	MOTATO,ACCESS		; no drive A timeout - keep it running
	andlw	0xFC				; mask drive select bits
	iorlw	0x10				; turn drive A motor on and select drive A

WriteDOR:
	movwf	DORVAL,ACCESS		; save current DOR value
	movwf	PORTA,ACCESS
	movlw	0x78				; /LDOR=0, A=0, /CS=1: select FDD controller operations register
	movwf	PORTC,ACCESS
	bcf		LATC,4,ACCESS		; /WRI activated (important: do not use PORTC as bcf is rmw operation)
	nop							; /WRI impulse width: 125ns (>60ns as per datasheet)
	bsf		LATC,4,ACCESS		; /WRI deactivated (important: do not use PORTC as bcf is rmw operation)
	setf	PORTC,ACCESS
	setf	PORTA,ACCESS

	btfsc	TMP,0,ACCESS
	return						; skip drive spin up delay if TMP<>0 (selected drive motor was already running)

	movlw	0xFF
	call	Wait_ms
	movlw	0xFF
	call	Wait_ms
	movlw	0xFF
	call	Wait_ms
	movlw	0xFF
	call	Wait_ms				; wait 255*4 ~ 1 second
	return
; ****************** Turn motor on and clear turoff counter ****************

; ------------------ Set up counter to turn motor off ----------------
; in:	CURDRV = current drive #
; corrupt: none
MotorOff:
	btfss	CURDRV,0,ACCESS
	bra		MotorCountDownDrv0

	setf	MOTBTO,ACCESS		; drive B timeout set
	return

MotorCountDownDrv0:
	setf	MOTATO,ACCESS		; drive A timeout set
	return
; ****************** Set up counter to turn motor off ****************



; ------------------ Clear data buffer  ---------------------
; in:	none
ClearBufForFormat:
	clrf	RECLENL,ACCESS
	movlw	0x02
	movwf	RECLENH,ACCESS
	movlw	0x01
	movwf	FSR0H,ACCESS
	clrf	FSR0L,ACCESS			; buffer address is at 0x100

ClearFillLoop:
	clrf	POSTINC0,ACCESS
	decfsz	RECLENL,1,ACCESS
	bra		ClearFillLoop
	decfsz	RECLENH,1,ACCESS		; increase byte count
	bra		ClearFillLoop
	return
; ****************** Clear data buffer  *********************

; ------------------ Fill data buffer with data for format 1.44MiB diskette ------------
; in:	W = cylinder to format, 18 sectors per track
; (!!!) Important: sector interleaving is NOT implemented
FillBufForFormat:
	movwf	DATAVAL,ACCESS
	clrf	RECLENL,ACCESS
	clrf	RECLENH,ACCESS
	movlw	0x01
	movwf	TMP,ACCESS				; sector number
	movwf	FSR0H,ACCESS
	clrf	FSR0L,ACCESS			; buffer address = 0x100

FormatFillCycle:
	movf	DATAVAL,W,ACCESS		; cylinder # (track)
	movwf	POSTINC0,ACCESS

	rrncf	CMDWLOC+1,W,ACCESS
	rrncf	WREG,W,ACCESS
	andlw	0x01
	movwf	POSTINC0,ACCESS			; head
	movf	TMP,W,ACCESS
	movwf	POSTINC0,ACCESS			; sector#
	movlw	0x02
	movwf	POSTINC0,ACCESS			; sector size (0x200)
	incf	TMP,1,ACCESS			; increase number of the sector

	movf	TMP,W,ACCESS
	sublw	0x13					; let routine fill all 18 sector space
	bnz		FormatFillCycle			; even if we format 9 sectors per track disk
	return
; ****************** Fill data buffer with data for format 1.44MiB diskette ************

;============================= Microcontroller <-> Host communication =========================
; ---------------- Waits for host write --------------------
; in:	none
; out:	WREG=byte read from the host; CF=1 if A0 was 1 (port base+1)
; corrupt:	LATD, STATUS
WaitForHostWrite:
	movf	PORTE,W,ACCESS
	movf	PORTD,1,ACCESS			; read port D value and write it into the LATD
	btfss	WREG,3,ACCESS
	bra		WaitForHostWrite

	rrcf	WREG,W,ACCESS			; shift A0 address bit (bit 0) to C
	movf	LATD,W,ACCESS			; get previously latched input data to WREG
	return

; ---------------- Waits for host read --------------------
; in:	WREG=byte to return to host
; out:	CF=0 if was read from port base+0, WREG is preserved!
; corrupt:	TRISD, LATD, STATUS
WaitForHostRead:
	movwf	PORTD,ACCESS			; write value to the port D output latch
									; PORTD is still functions as input
WaitWhilePICRDInactive:
	btfsc	PORTE,2,ACCESS
	bra		WaitWhilePICRDInactive	; Loop while /PICRD is active (skip this host read
									; cycle as its beginning is already missed)
WaitForPICRDAct:
	btfss	PORTE,2,ACCESS
	bra		WaitForPICRDAct			; Waiting for activation of /PICRD

	; /PICRD was activated - outputting port D
	clrf	TRISD,ACCESS

	; get address MC accessed with
	rrcf	PORTE,W,ACCESS			; "A0" (bit 0) -> CF

	; now watitng for /PICRD to deactivation
WaitForPICRDDeact:
	btfsc	PORTE,2,ACCESS
	bra		WaitForPICRDDeact		; Waiting for de-activation of /PICRD

	setf	TRISD,ACCESS
	movf	LATD,W,ACCESS			; restore initial value of WREG, does not affect C flag
	return
;============================= End of "Microcontroller <-> Host communication" =========================

;============================= Check for single FDD configuration =========================
; needed to operate phantom drive 1
; returns CURDRV=_physical_ drive to access
CheckSingleDrive:
; in:	_logical_ drive number requested
	movwf	CURHOSTDRV,ACCESS		; updated currently host selected drive
	movwf	CURDRV,ACCESS
	movf	DRINUM,W,ACCESS
	sublw	0x01					; is there only one drive?
	btfsc	STATUS,Z,ACCESS
	clrf	CURDRV,ACCESS			; current drive = drive 0 (while it can be phantom drive B: for host)
	return

;============================= Microcontroller interrupt handler =========================
; is used to turn off drives' motors. Is only executed when interrupts are enabled
int_handler:
	; if MOTxTO is not 0, decrement it and when it is zero stop the motor but not deselect drive
	movf	MOTATO,1,ACCESS			; check Z flag
	btfsc	STATUS,Z,ACCESS
	bra		StopDrvB				; MOTATO=0, do not count it down

	decfsz	MOTATO,1,ACCESS
	bra		StopDrvB				; counter did not reach 0 for drive A

	movf	DORVAL,W,ACCESS
	andlw	0xEF					; bit 4 reset
	call	WriteDOR

StopDrvB:
	movf	MOTBTO,1,ACCESS			; check Z flag
	btfsc	STATUS,Z,ACCESS
	bra		RetInt					; MOTBTO=0, do not count it down

	decfsz	MOTBTO,1,ACCESS
	bra		RetInt					; counter did not reach 0 for drive B

	movf	DORVAL,W,ACCESS
	andlw	0xDF					; bit 5 reset
	call	WriteDOR

RetInt:
	clrf	TMR1L,ACCESS
	clrf	TMR1H,ACCESS
	clrf	PIR1,ACCESS				; clear interrupt condition flag
	retfie	FAST					; return from interrupt and restore main registers

#include<hdd_access_v2_0a.inc>


	END